'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Union, Tuple, List

import numpy as np
import polars as pl

from pm4py.statistics.traces.generic.common import (
    case_duration as case_duration_commons,
)
from pm4py.util import exec_utils, constants
from pm4py.util import xes_constants as xes
from pm4py.util.business_hours import soj_time_business_hours_diff
from pm4py.util.constants import CASE_CONCEPT_NAME
from pm4py.util.xes_constants import DEFAULT_TIMESTAMP_KEY
from collections import Counter
from statistics import median


class Parameters(Enum):
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY

    MAX_VARIANTS_TO_RETURN = "max_variants_to_return"
    VARIANTS_DF = "variants_df"
    ENABLE_SORT = "enable_sort"
    SORT_BY_COLUMN = "sort_by_column"
    SORT_ASCENDING = "sort_ascending"
    MAX_RET_CASES = "max_ret_cases"

    BUSINESS_HOURS = "business_hours"
    BUSINESS_HOUR_SLOTS = "business_hour_slots"
    WORKCALENDAR = "workcalendar"


def get_variant_statistics(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> List[Dict[str, Any]]:
    """
    Get variants from a Polars LazyFrame

    Parameters
    -----------
    lf
        LazyFrame
    parameters
        Parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Column that contains the Case ID
            Parameters.ACTIVITY_KEY -> Column that contains the activity
            Parameters.MAX_VARIANTS_TO_RETURN -> Maximum number of variants to return

    Returns
    -----------
    variants_list
        List of variants inside the LazyFrame
    """
    if parameters is None:
        parameters = {}
    
    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )
    max_variants_to_return = exec_utils.get_param_value(
        Parameters.MAX_VARIANTS_TO_RETURN, parameters, None
    )

    # Get variants by grouping activities per case
    variants_df = (
        lf.group_by(case_id_glue, maintain_order=True)
        .agg(pl.col(activity_key).alias("variant"))
        .collect()
    )
    
    # Count occurrences of each variant
    variant_counts = Counter()
    for row in variants_df.iter_rows():
        variant = tuple(row[1])  # Convert list to tuple for hashing
        variant_counts[variant] += 1
    
    # Convert to list format
    variants_list = [
        {"variant": variant, case_id_glue: count}
        for variant, count in variant_counts.items()
    ]
    
    # Sort by count and variant
    variants_list = sorted(
        variants_list,
        key=lambda x: (x[case_id_glue], x["variant"]),
        reverse=True,
    )

    if max_variants_to_return:
        variants_list = variants_list[
            : min(len(variants_list), max_variants_to_return)
        ]
    
    return variants_list


def get_variants_df_and_list(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[pl.DataFrame, List[List[Any]]]:
    """
    (Technical method) Provides variants_df and variants_list out of the box

    Parameters
    ------------
    lf
        LazyFrame
    parameters
        Parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Column that contains the Case ID
            Parameters.ACTIVITY_KEY -> Column that contains the activity

    Returns
    ------------
    variants_df
        Variants dataframe
    variants_list
        List of variants sorted by their count
    """
    if parameters is None:
        parameters = {}
    
    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )

    variants_df = get_variants_df(lf, parameters=parameters)
    variants_stats = get_variant_statistics(lf, parameters=parameters)
    
    variants_list = []
    for vd in variants_stats:
        variant = vd["variant"]
        count = vd[case_id_glue]
        variants_list.append([variant, count])
    
    variants_list = sorted(
        variants_list, key=lambda x: (x[1], x[0]), reverse=True
    )
    
    return variants_df, variants_list


def get_cases_description(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> List[Dict[str, Any]]:
    """
    Get a description of traces in the dataframe

    Parameters
    -----------
    lf
        LazyFrame
    parameters
        Parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Column that contains the Case ID
            Parameters.TIMESTAMP_KEY -> Column that contains the timestamp

    Returns
    -----------
    ret
        Dictionary of cases with their start time, end time, and case duration
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )

    # Get case statistics
    case_stats = (
        lf.group_by(case_id_glue)
        .agg([
            pl.col(timestamp_key).min().alias("startTime"),
            pl.col(timestamp_key).max().alias("endTime")
        ])
        .with_columns(
            (pl.col("endTime") - pl.col("startTime")).dt.total_seconds().alias("caseDuration")
        )
        .collect()
    )

    ret = []
    for row in case_stats.iter_rows():
        case_id = row[0]
        start_time = row[1]
        end_time = row[2]
        duration = row[3]
        
        ret.append({
            "caseid": case_id,
            "startTime": start_time,
            "endTime": end_time,
            "caseDuration": duration
        })

    return ret


def get_variants_df(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.DataFrame:
    """
    Gets the variants dataframe

    Parameters
    -----------
    lf
        LazyFrame
    parameters
        Parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Column that contains the Case ID
            Parameters.ACTIVITY_KEY -> Column that contains the activity

    Returns
    -----------
    variants_df
        Variants dataframe
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )

    # Create variants dataframe
    variants_df = (
        lf.group_by(case_id_glue, maintain_order=True)
        .agg(pl.col(activity_key).alias("variant"))
        .collect()
    )

    return variants_df


def get_all_case_durations(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> List[float]:
    """
    Gets all case durations

    Parameters
    -----------
    lf
        LazyFrame
    parameters
        Parameters of the algorithm

    Returns
    -----------
    case_durations
        List of case durations
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )

    # Get case durations
    durations_df = (
        lf.group_by(case_id_glue)
        .agg([
            pl.col(timestamp_key).min().alias("start_time"),
            pl.col(timestamp_key).max().alias("end_time")
        ])
        .with_columns(
            (pl.col("end_time") - pl.col("start_time")).dt.total_seconds().alias("duration")
        )
        .collect()
    )

    return durations_df["duration"].to_list()


def get_median_case_duration(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> float:
    """
    Gets the median case duration

    Parameters
    -----------
    lf
        LazyFrame
    parameters
        Parameters of the algorithm

    Returns
    -----------
    median_case_duration
        Median case duration
    """
    durations = get_all_case_durations(lf, parameters)
    return median(durations)


def get_first_quartile_case_duration(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> float:
    """
    Gets the first quartile case duration

    Parameters
    -----------
    lf
        LazyFrame
    parameters
        Parameters of the algorithm

    Returns
    -----------
    first_quartile_case_duration
        First quartile case duration
    """
    durations = get_all_case_durations(lf, parameters)
    return float(np.percentile(durations, 25))


def get_kde_caseduration(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[List[float], List[float]]:
    """
    Gets the KDE estimation for case durations

    Parameters
    -----------
    lf
        LazyFrame
    parameters
        Parameters of the algorithm

    Returns
    -----------
    x, y
        X and Y values for KDE
    """
    durations = get_all_case_durations(lf, parameters)
    return case_duration_commons.get_kde_caseduration(durations, parameters)


def get_kde_caseduration_json(
    lf: pl.LazyFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, Any]:
    """
    Gets the KDE estimation for case durations (JSON format)

    Parameters
    -----------
    lf
        LazyFrame
    parameters
        Parameters of the algorithm

    Returns
    -----------
    json_data
        JSON representation of KDE
    """
    durations = get_all_case_durations(lf, parameters)
    return case_duration_commons.get_kde_caseduration_json(durations, parameters)


'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import difflib
from typing import List, Dict, Set


# ---------------------------------------------------------------------------
# 1.  Similarity between two (unordered) label collections
# ---------------------------------------------------------------------------

def label_sets_similarity(
        labels_a: List[str],
        labels_b: List[str],
        threshold: float = 0.75,
) -> float:
    """
    Dice-style similarity ∈ [0,1] that tolerates near-matches.

    * Each label in A is greedily paired with its most similar,
      still-unmatched label in B (SequenceMatcher ratio ≥ `threshold`).
    * Similarity  =  2·|matched| / (|A| + |B|).
    """
    remaining_b: Set[str] = set(labels_b)
    matches = 0

    for la in labels_a:
        if not remaining_b:  # B exhausted
            break
        # find best candidate among the still-free labels in B
        best, score = None, 0.0
        for lb in remaining_b:
            r = difflib.SequenceMatcher(None, la, lb).ratio()
            if r > score:
                best, score = lb, r
        if score >= threshold:
            matches += 1
            remaining_b.remove(best)

    return 2 * matches / (len(labels_a) + len(labels_b)) if (labels_a or labels_b) else 1.0


# ---------------------------------------------------------------------------
# 2.  Greedy label mapping A → B (or A → A if no good match)
# ---------------------------------------------------------------------------

def map_labels(
        labels_a: List[str],
        labels_b: List[str],
        threshold: float = 0.75,
) -> Dict[str, str]:
    """
    Returns a dict {label_in_A: mapped_label}.

    * If an exact duplicate exists in B, map to that.
    * Otherwise pick the *unused* label in B with the highest similarity
      (SequenceMatcher ratio ≥ `threshold`).
    * If nothing reaches the threshold, map to itself (no change).
    """
    mapping: Dict[str, str] = {}
    used_b: Set[str] = set()

    for la in labels_a:
        # 1) perfect match first
        if la in labels_b and la not in used_b:
            mapping[la] = la
            used_b.add(la)
            continue

        # 2) best fuzzy match
        best, score = None, 0.0
        for lb in labels_b:
            if lb in used_b:
                continue
            r = difflib.SequenceMatcher(None, la, lb).ratio()
            if r > score:
                best, score = lb, r

        if best is not None and score >= threshold:
            mapping[la] = best
            used_b.add(best)
        else:
            mapping[la] = la  # fallback: keep original

    return mapping



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
"""Polars implementation for extracting DFG frequency triples."""

from typing import Dict, Tuple

import polars as pl


def get_freq_triples(
    df: pl.LazyFrame,
    activity_key: str = "concept:name",
    case_id_glue: str = "case:concept:name",
    timestamp_key: str = "time:timestamp",
    sort_caseid_required: bool = True,
    sort_timestamp_along_case_id: bool = True,
) -> Dict[Tuple[str, str, str], int]:
    """Compute frequency triples directly on a Polars LazyFrame."""

    if sort_caseid_required:
        if sort_timestamp_along_case_id:
            df = df.sort([case_id_glue, timestamp_key])
        else:
            df = df.sort(case_id_glue)

    triples = df.with_columns(
        [
            pl.col(activity_key)
            .shift(-1)
            .over(case_id_glue)
            .alias(activity_key + "_2"),
            pl.col(activity_key)
            .shift(-2)
            .over(case_id_glue)
            .alias(activity_key + "_3"),
        ]
    )

    triples = triples.filter(
        pl.col(activity_key + "_2").is_not_null()
        & pl.col(activity_key + "_3").is_not_null()
    )

    grouped = (
        triples.group_by(
            [activity_key, activity_key + "_2", activity_key + "_3"]
        )
        .agg(pl.len().alias("count"))
        .collect()
    )

    return {
        (
            row[activity_key],
            row[activity_key + "_2"],
            row[activity_key + "_3"],
        ): row["count"]
        for row in grouped.iter_rows(named=True)
    }



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME
from pm4py.util import xes_constants as xes
from pm4py.util.xes_constants import DEFAULT_NAME_KEY
from pm4py.util.constants import PARAMETER_CONSTANT_CASEID_KEY, PARAMETER_CONSTANT_ACTIVITY_KEY
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union, List
import polars as pl


class Parameters(Enum):
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    POSITIVE = "positive"


def apply(
    df: pl.LazyFrame,
    values: List[str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Filter LazyFrame on start activities

    Parameters
    ----------
    df
        LazyFrame
    values
        Values to filter on
    parameters
        Parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Case ID column
            Parameters.ACTIVITY_KEY -> Activity column
            Parameters.POSITIVE -> Include traces (True) or exclude traces (False)

    Returns
    ----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(Parameters.POSITIVE, parameters, True)

    # Get cases with matching start activities
    matching_cases = (
        df.group_by(case_id_glue)
        .agg(pl.col(activity_key).first().alias("start_activity"))
        .filter(pl.col("start_activity").is_in(values))
        .select(case_id_glue)
    )
    
    if positive:
        ret = df.join(matching_cases, on=case_id_glue, how="inner")
    else:
        ret = df.join(matching_cases, on=case_id_glue, how="anti")

    return ret



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import sys
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any
import base64
import os
from pm4py.util import constants


class Parameters(Enum):
    API_URL = "api_url"
    API_KEY = "api_key"
    OPENAI_MODEL = "openai_model"
    IMAGE_PATH = "image_path"
    MAX_TOKENS = "max_tokens"
    USE_RESPONSES_API = "use_responses_api"
    EXTRA_PAYLOAD = "extra_payload"


def encode_image(image_path):
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")


def apply(prompt: str, parameters: Optional[Dict[Any, Any]] = None) -> str:
    import requests

    if parameters is None:
        parameters = {}

    image_path = exec_utils.get_param_value(
        Parameters.IMAGE_PATH, parameters, None
    )
    api_key = exec_utils.get_param_value(
        Parameters.API_KEY, parameters, constants.OPENAI_API_KEY
    )
    api_url = exec_utils.get_param_value(Parameters.API_URL, parameters, None)
    simple_content_specification = image_path is None
    max_tokens = exec_utils.get_param_value(
        Parameters.MAX_TOKENS, parameters, None
    )
    extra_payload = exec_utils.get_param_value(
        Parameters.EXTRA_PAYLOAD, parameters, {}
    )

    if api_url is None:
        api_url = constants.OPENAI_API_URL
    else:
        if not api_url.endswith("/"):
            api_url += "/"

    use_responses_api = exec_utils.get_param_value(Parameters.USE_RESPONSES_API, parameters, "api.openai" in api_url)

    model = exec_utils.get_param_value(
        Parameters.OPENAI_MODEL,
        parameters,
        (
            constants.OPENAI_DEFAULT_MODEL
            if image_path is None
            else constants.OPENAI_DEFAULT_VISION_MODEL
        ),
    )

    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {api_key}",
    }

    messages = []
    payload = {"model": model}

    if use_responses_api:
        messages.append(
            {"role": "user", "content": [{"type": "input_text", "text": prompt}]}
        )
    else:
        if simple_content_specification:
            messages.append({"role": "user", "content": prompt})
        else:
            messages.append(
                {"role": "user", "content": [{"type": "text", "text": prompt}]}
            )

    if image_path is not None:
        max_tokens = exec_utils.get_param_value(
            Parameters.MAX_TOKENS, parameters, 16384
        )
        image_format = os.path.splitext(image_path)[1][1:].lower()
        base64_image = encode_image(image_path)

        if use_responses_api:
            messages[0]["content"].append({
                "type": "input_image",
                "image_url": f"data:image/{image_format};base64,{base64_image}"
            })
        else:
            messages[0]["content"].append(
                {
                    "type": "image_url",
                    "image_url": {
                        "url": f"data:image/{image_format};base64,{base64_image}"
                    },
                }
            )
            payload["max_tokens"] = max_tokens

    if max_tokens is not None and not use_responses_api:
        payload["max_tokens"] = max_tokens

    if use_responses_api:
        payload["input"] = messages
    else:
        payload["messages"] = messages

    if extra_payload:
        payload.update(extra_payload)

    if use_responses_api:
        response = requests.post(
            api_url + "responses", headers=headers, json=payload, timeout=20*60
        )

        response = response.json()

        if "error" in response and response["error"]:
            # raise an exception when the request fails, with the provided message
            raise Exception(response["error"]["message"])

        return response["output"][-1]["content"][0]["text"]
    else:
        response = requests.post(
            api_url + "chat/completions", headers=headers, json=payload, timeout=20*60
        ).json()

        if "error" in response:
            # raise an exception when the request fails, with the provided message
            raise Exception(response["error"]["message"])

        return response["choices"][0]["message"]["content"]



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.log.obj import EventLog, XESExtension
from typing import Optional, Dict, Any, Union
from enum import Enum
from pm4py.util import exec_utils
from copy import deepcopy


class Parameters(Enum):
    ENABLE_DEEPCOPY = "enable_deepcopy"


def apply(
    log: EventLog,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Moves the attributes that are constant for all the events of the trace, and they
    do not belong to a standard extension, to the trace level

    Parameters
    ----------------
    log
        Event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.DEEPCOPY => enables the deepcopy of the event log

    Returns
    ----------------
    log
        Event log, where some attribute has been possibly moved from the event to the trace level
    """
    if parameters is None:
        parameters = {}

    enable_deepcopy = exec_utils.get_param_value(
        Parameters.ENABLE_DEEPCOPY, parameters, False
    )
    if enable_deepcopy:
        log = deepcopy(log)

    main_extensions = set()
    for e in XESExtension:
        main_extensions.add(e.value[1])

    candidates = None
    for trace in log:
        values_count = {}
        for eve in trace:
            for attr in eve:
                if attr.split(":")[0] not in main_extensions:
                    if attr not in trace.attributes:
                        if attr not in values_count:
                            values_count[attr] = [eve[attr]]
                        else:
                            values_count[attr].append(eve[attr])
        trace_candidates = set(
            x
            for x in values_count
            if len(values_count[x]) == len(trace)
            and len(set(values_count[x])) == 1
        )
        if candidates is not None:
            candidates = trace_candidates.intersection(candidates)
        else:
            candidates = trace_candidates

    for attr in candidates:
        for trace in log:
            trace.attributes[attr] = trace[0][attr]
            for ev in trace:
                del ev[attr]

    return log



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any
from enum import Enum
from pm4py.util import exec_utils, pandas_utils
from pm4py.objects.ocel import constants as ocel_constants
from pm4py.objects.ocel.obj import OCEL
from pm4py.statistics.ocel import act_ot_dependent, act_utils, edge_metrics


class Parameters(Enum):
    EVENT_ACTIVITY = ocel_constants.PARAM_EVENT_ACTIVITY
    OBJECT_TYPE = ocel_constants.PARAM_OBJECT_TYPE
    COMPUTE_EDGES_PERFORMANCE = "compute_edges_performance"


def apply(
    ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> Dict[str, Any]:
    """
    Discovers an OC-DFG model from an object-centric event log.
    Reference paper:
    Berti, Alessandro, and Wil van der Aalst. "Extracting multiple viewpoint models from relational databases." Data-Driven Process Discovery and Analysis. Springer, Cham, 2018. 24-51.

    Parameters
    -----------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ACTIVITY => the attribute to be used as activity
        - Parameters.OBJECT_TYPE => the attribute to be used as object type
        - Parameters.COMPUTE_EDGES_PERFORMANCE => (boolean) enables/disables the computation of the performance on the edges

    Returns
    -----------------
    ocdfg
        Object-centric directly-follows graph, expressed as a dictionary containing the following properties:
        - activities: complete set of activities derived from the object-centric event log
        - object_types: complete set of object types derived from the object-centric event log
        - edges: dictionary connecting each object type to a set of directly-followed arcs between activities
        - activities_indep: dictionary linking each activity, regardless of the object type
        - activities_ot: dictionary linking each object type to another dictionary
        - start_activities: dictionary linking each object type to start activities
        - end_activities: dictionary linking each object type to end activities
    """
    if parameters is None:
        parameters = {}

    # Extract parameter values once
    object_type = exec_utils.get_param_value(
        Parameters.OBJECT_TYPE, parameters, ocel.object_type_column
    )
    event_activity = exec_utils.get_param_value(
        Parameters.EVENT_ACTIVITY, parameters, ocel.event_activity
    )
    compute_edges_performance = exec_utils.get_param_value(
        Parameters.COMPUTE_EDGES_PERFORMANCE, parameters, True
    )

    # Pre-compute activities and object types (used multiple times)
    activities = set(pandas_utils.format_unique(ocel.events[event_activity].unique()))
    object_types = set(pandas_utils.format_unique(ocel.objects[object_type].unique()))

    # Initialize result dictionary with pre-computed values
    ret = {
        "activities": activities,
        "object_types": object_types,
        "edges": {},
        "activities_indep": {},
        "activities_ot": {},
        "start_activities": {},
        "end_activities": {},
        "edges_performance": {"event_couples": {}, "total_objects": {}}
    }

    # Process object-type independent associations (shared computation)
    ot_independent = act_utils.find_associations_from_ocel(
        ocel, parameters=parameters
    )
    ret["activities_indep"] = {
        "events": act_utils.aggregate_events(ot_independent),
        "unique_objects": act_utils.aggregate_unique_objects(ot_independent),
        "total_objects": act_utils.aggregate_total_objects(ot_independent)
    }

    # Process object-type dependent associations (shared computation)
    ot_dependent = act_ot_dependent.find_associations_from_ocel(
        ocel, parameters=parameters
    )
    ret["activities_ot"] = {
        "events": act_ot_dependent.aggregate_events(ot_dependent),
        "unique_objects": act_ot_dependent.aggregate_unique_objects(ot_dependent),
        "total_objects": act_ot_dependent.aggregate_total_objects(ot_dependent)
    }

    # Process start activities
    start_parameters = {**parameters, "prefiltering": "start"}
    ot_dependent_start = act_ot_dependent.find_associations_from_ocel(
        ocel, parameters=start_parameters
    )
    ret["start_activities"] = {
        "events": act_ot_dependent.aggregate_events(ot_dependent_start),
        "unique_objects": act_ot_dependent.aggregate_unique_objects(ot_dependent_start),
        "total_objects": act_ot_dependent.aggregate_total_objects(ot_dependent_start)
    }

    # Process end activities
    end_parameters = {**parameters, "prefiltering": "end"}
    ot_dependent_end = act_ot_dependent.find_associations_from_ocel(
        ocel, parameters=end_parameters
    )
    ret["end_activities"] = {
        "events": act_ot_dependent.aggregate_events(ot_dependent_end),
        "unique_objects": act_ot_dependent.aggregate_unique_objects(ot_dependent_end),
        "total_objects": act_ot_dependent.aggregate_total_objects(ot_dependent_end)
    }

    # Process edges
    edges = edge_metrics.find_associations_per_edge(
        ocel, parameters=end_parameters
    )
    ret["edges"] = {
        "event_couples": edge_metrics.aggregate_ev_couples(edges),
        "unique_objects": edge_metrics.aggregate_unique_objects(edges),
        "total_objects": edge_metrics.aggregate_total_objects(edges)
    }

    # Only compute performance metrics if needed
    if compute_edges_performance:
        ret["edges_performance"] = {
            "event_couples": edge_metrics.performance_calculation_ocel_aggregation(
                ocel, ret["edges"]["event_couples"], parameters=parameters
            ),
            "total_objects": edge_metrics.performance_calculation_ocel_aggregation(
                ocel, ret["edges"]["total_objects"], parameters=parameters
            )
        }

    return ret



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
class Edge:
    def __init__(
        self,
        start_node,
        end_node,
        dependency_value,
        dfg_value,
        repr_value,
        label="",
        repr_color="#000000",
        edge_type="frequency",
        net_name="",
    ):
        """
        Constructor

        Parameters
        ------------
        start_node
            Start node of the edge
        end_node
            End node of the edge
        dependency_value
            Dependency value of the edge
        dfg_value
            DFG value of the edge
        repr_value
            Value used in the representation
        label
            (If provided) label of the edge
        repr_color
            (If provided) Color that shall represent the edge
        edge_type
            Type of the edge (frequency or performance)
        net_name
            (If provided) name of the Heuristics Net
        """
        self.start_node = start_node
        self.end_node = end_node
        self.dependency_value = dependency_value
        self.dfg_value = dfg_value
        self.repr_value = repr_value
        self.label = label
        self.repr_color = repr_color
        self.edge_type = edge_type
        self.net_name = net_name
        self.repr_font_color = None
        self.pen_width = None

    def get_color(self):
        """
        Gets the color to use for the representation
        """
        return self.repr_color

    def get_font_color(self):
        """
        Gets the font color to use for the representation
        """
        if self.repr_font_color is not None:
            return self.repr_font_color
        return self.repr_color

    def get_penwidth(self, default):
        """
        Gets the pen width to use for the representation

        Parameters
        --------------
        default
            Default value
        """
        if self.pen_width is not None:
            return self.pen_width
        return default



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME
from pm4py.statistics.traces.generic.pandas.case_statistics import (
    get_variants_df,
)
from pm4py.statistics.variants.pandas import get as variants_get
from pm4py.util.constants import (
    PARAMETER_CONSTANT_CASEID_KEY,
    PARAMETER_CONSTANT_ACTIVITY_KEY,
)
from enum import Enum
from pm4py.util import exec_utils
from copy import copy
from typing import Optional, Dict, Any, Union, List
import pandas as pd


class Parameters(Enum):
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    DECREASING_FACTOR = "decreasingFactor"
    POSITIVE = "positive"


def apply(
    df: pd.DataFrame,
    admitted_variants: List[List[str]],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pd.DataFrame:
    """
    Apply a filter on variants

    Parameters
    -----------
    df
        Dataframe
    admitted_variants
        List of admitted variants (to include/exclude)
    parameters
        Parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Column that contains the Case ID
            Parameters.ACTIVITY_KEY -> Column that contains the activity
            Parameters.POSITIVE -> Specifies if the filter should be applied including traces (positive=True)
            or excluding traces (positive=False)
            variants_df -> If provided, avoid recalculation of the variants dataframe

    Returns
    -----------
    df
        Filtered dataframe
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )
    variants_df = (
        parameters["variants_df"]
        if "variants_df" in parameters
        else get_variants_df(df, parameters=parameters)
    )
    variants_df = variants_df[variants_df["variant"].isin(admitted_variants)]
    i1 = df.set_index(case_id_glue).index
    i2 = variants_df.index
    if positive:
        ret = df[i1.isin(i2)]
    else:
        ret = df[~i1.isin(i2)]

    ret.attrs = copy(df.attrs) if hasattr(df, "attrs") else {}
    return ret


def filter_variants_top_k(log, k, parameters=None):
    """
    Keeps the top-k variants of the log

    Parameters
    -------------
    log
        Event log
    k
        Number of variants that should be kept
    parameters
        Parameters

    Returns
    -------------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    variants = variants_get.get_variants_count(log, parameters=parameters)
    variant_count = []
    for variant in variants:
        variant_count.append([variant, variants[variant]])
    variant_count = sorted(
        variant_count, key=lambda x: (x[1], x[0]), reverse=True
    )
    variant_count = variant_count[: min(k, len(variant_count))]
    variants_to_filter = [x[0] for x in variant_count]

    return apply(log, variants_to_filter, parameters=parameters)


def filter_variants_by_coverage_percentage(
    log, min_coverage_percentage, parameters=None
):
    """
    Filters the variants of the log by a coverage percentage
    (e.g., if min_coverage_percentage=0.4, and we have a log with 1000 cases,
    of which 500 of the variant 1, 400 of the variant 2, and 100 of the variant 3,
    the filter keeps only the traces of variant 1 and variant 2).

    Parameters
    ---------------
    log
        Event log
    min_coverage_percentage
        Minimum allowed percentage of coverage
    parameters
        Parameters

    Returns
    ---------------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )

    variants = variants_get.get_variants_count(log, parameters=parameters)
    allowed_variants = [
        x
        for x, y in variants.items()
        if y >= min_coverage_percentage * log[case_id_glue].nunique()
    ]

    return apply(log, allowed_variants, parameters=parameters)


def filter_variants_by_maximum_coverage_percentage(log, max_coverage_percentage, parameters=None):
    """
    Filters the variants of the log by a maximum coverage percentage
    (e.g., if max_coverage_percentage=0.4, and we have a log with 1000 cases,
    of which 500 of the variant 1, 400 of the variant 2, and 100 of the variant 3,
    the filter keeps only the traces of variant w and variant 3).

    Parameters
    ---------------
    log
        Event log
    max_coverage_percentage
        Maximum allowed percentage of coverage
    parameters
        Parameters

    Returns
    ---------------
    filtered_log
        Filtered log
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME)

    variants = variants_get.get_variants_count(log, parameters=parameters)
    allowed_variants = [x for x, y in variants.items() if y <= max_coverage_percentage * log[case_id_glue].nunique()]

    return apply(log, allowed_variants, parameters=parameters)


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.algo.transformation.log_to_features import (
    algorithm as features_extractor,
)
from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd
import numpy as np
from typing import Optional, Dict, Any, Generator, Union
from copy import copy


class Parameters(Enum):
    SKLEARN_CLUSTERER = "sklearn_clusterer"


def apply(
    log: Union[EventLog, EventStream, pd.DataFrame],
    parameters: Optional[Dict[Any, Any]] = None,
) -> Generator[EventLog, None, None]:
    """
    Cluster the event log, based on the extraction of profiles for the traces of the event log
    (by means of the feature extraction proposed in pm4py) and the application of a Scikit learn clusterer
    (default: K-means with two clusters)

    Implements the approach described in:
    Song, Minseok, Christian W. Günther, and Wil MP Van der Aalst. "Trace clustering in process mining." Business Process Management Workshops: BPM 2008 International Workshops, Milano, Italy, September 1-4, 2008. Revised Papers 6. Springer Berlin Heidelberg, 2009.

    Parameters
    ----------------
    log
        Event log
    parameters
        Parameters of the feature extraction, including:
        - Parameters.SKLEARN_CLUSTERER => the Scikit-Learn clusterer to be used (default: KMeans(n_clusters=2, random_state=0, n_init="auto"))

    Returns
    ---------------
    generator
        Generator of logs (clusters)
    """
    if parameters is None:
        parameters = {}

    from pm4py.util import ml_utils

    clusterer = exec_utils.get_param_value(
        Parameters.SKLEARN_CLUSTERER,
        parameters,
        ml_utils.KMeans(n_clusters=2, random_state=0, n_init="auto"),
    )

    if "enable_activity_def_representation" not in parameters:
        parameters["enable_activity_def_representation"] = True

    if "enable_succ_def_representation" not in parameters:
        parameters["enable_succ_def_representation"] = True

    conv_parameters = copy(parameters)
    conv_parameters["stream_postprocessing"] = True

    log = log_converter.apply(
        log,
        variant=log_converter.Variants.TO_EVENT_LOG,
        parameters=conv_parameters,
    )
    data, feature_names = features_extractor.apply(log, parameters=parameters)
    data = np.array([np.array(x) for x in data])

    clusters = clusterer.fit_predict(data)
    max_clu = max(clusters)
    clust_idxs = {i: list() for i in range(max_clu + 1)}

    for i in range(len(clusters)):
        clust_idxs[clusters[i]].append(i)

    for i in clust_idxs:
        clust_log = EventLog()
        for j in clust_idxs[i]:
            clust_log.append(log[j])
        # clust_log = log_converter.apply(clust_log, variant=log_converter.Variants.TO_DATA_FRAME, parameters=parameters)

        yield clust_log



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from pm4py.util import exec_utils, constants
from enum import Enum
import numpy as np
import pandas as pd


class Parameters(Enum):
    INCLUDE_HEADER = "include_header"
    MAX_LEN = "max_len"
    DEBUG = "debug"
    ENABLE_OBJECT_LIFECYCLE_PATHS = "enable_object_lifecycle_paths"


def __transform_to_string(stru: str) -> str:
    if stru.startswith("@@ocel_lif_activity_"):
        return (
            "Number of occurrences of the activity "
            + stru.split("@@ocel_lif_activity_")[1]
        )
    elif stru.startswith("@@object_lifecycle_unq_act"):
        return "Number of unique activities in the lifecycle of the object"
    elif stru.startswith("@@object_lifecycle_length"):
        return "Number of events in the lifecycle of the object"
    elif stru.startswith("@@object_lifecycle_duration"):
        return "Duration of the lifecycle of the object"
    elif stru.startswith("@@object_lifecycle_start_timestamp"):
        return "Start timestamp of the lifecycle of the object"
    elif stru.startswith("@@object_lifecycle_end_timestamp"):
        return "Completion timestamp of the lifecycle of the object"
    elif stru.startswith("@@object_degree_centrality"):
        return (
            "Degree centrality of the object in the object interaction graph"
        )
    elif stru.startswith("@@object_general_interaction_graph"):
        return "Number of objects related in the object interaction graph"
    elif stru.startswith("@@object_general_descendants_graph_descendants"):
        return "Number of objects which follow the current object in the object descendants graph"
    elif stru.startswith("@@object_general_inheritance_graph_ascendants"):
        return "Number of objects which follow the current object in the object inheritance graph"
    elif stru.startswith("@@object_general_descendants_graph_ascendants"):
        return "Number of objects which precede the current object in the object descendants graph"
    elif stru.startswith("@@object_general_inheritance_graph_descendants"):
        return "Number of objects which precede the current object in the object descendants graph"
    elif stru.startswith("@@object_cobirth"):
        return "Number of objects starting their lifecycle together with the current object"
    elif stru.startswith("@@object_codeath"):
        return "Number of objects ending their lifecycle together with the current object"
    elif stru.startswith("@@object_interaction_graph_"):
        return (
            "Number of object of type "
            + stru.split("@@object_interaction_graph_")[1]
            + " related to the current object in the object interaction graph"
        )
    elif stru.startswith("@@ocel_lif_path_"):
        path = stru.split("@@ocel_lif_path_")[1]
        act1 = path.split("##")[0]
        act2 = path.split("##")[1]
        return (
            'Frequency of the path "'
            + act1
            + '" -> "'
            + act2
            + '" in the lifecycle of the object'
        )

    return stru


def textual_abstraction_from_fea_df(
    fea_df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
) -> str:
    if parameters is None:
        parameters = {}

    include_header = exec_utils.get_param_value(
        Parameters.INCLUDE_HEADER, parameters, True
    )
    max_len = exec_utils.get_param_value(
        Parameters.MAX_LEN, parameters, constants.OPENAI_MAX_LEN
    )

    cols = []

    for c in fea_df.columns:
        ser = fea_df[c]
        ser1 = ser[ser > 0]
        if len(ser1) > 0:
            desc = __transform_to_string(c)
            avg = np.average(ser1)
            stdavg = 0 if avg == 0 or len(ser1) == 1 else np.std(ser1) / avg
            cols.append([desc, len(ser1), stdavg, ser1])

    cols = sorted(cols, key=lambda x: (x[1], x[2], x[0]), reverse=True)

    ret = ["\n"]

    if include_header:
        ret.append("Beforehand, a bit of notions.")
        ret.append(
            "Given an object-centric event log, the object interaction graph connects objects that are related in at least an event."
        )
        ret.append(
            "The object descendants graph connects objects related in at least an event, when the lifecycle of the second object starts after the lifecycle of the first."
        )
        ret.append(
            "The object inheritance graph connects objects when there an event that ends the lifecycle of the first object and starts the lifecycle of the second one."
        )
        ret.append("\n\n")
        ret.append("Given the following features:\n\n")

    ret = " ".join(ret)

    i = 0
    while i < len(cols):
        if len(ret) >= max_len:
            break

        stru = (
            cols[i][0]
            + ":    number of non-zero values: "
            + str(cols[i][1])
            + " ; quantiles of the non-zero: "
            + str(cols[i][3].quantile([0.0, 0.25, 0.5, 0.75, 1.0]).to_dict())
            + "\n"
        )
        ret = ret + stru

        i = i + 1

    return ret


def apply(
    ocel: OCEL, obj_type: str, parameters: Optional[Dict[Any, Any]] = None
) -> str:
    if parameters is None:
        parameters = {}

    debug = exec_utils.get_param_value(Parameters.DEBUG, parameters, True)
    enable_object_lifecycle_paths = exec_utils.get_param_value(
        Parameters.ENABLE_OBJECT_LIFECYCLE_PATHS, parameters, False
    )

    import pm4py

    fea_df = pm4py.extract_ocel_features(
        ocel,
        obj_type,
        include_obj_id=False,
        debug=debug,
        enable_object_lifecycle_paths=enable_object_lifecycle_paths,
    )

    return textual_abstraction_from_fea_df(fea_df, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils
from pm4py.util import xes_constants, constants, pandas_utils
import pandas as pd
from typing import Dict, Optional, Any, Tuple
from pm4py.util.business_hours import soj_time_business_hours_diff
from pm4py.algo.discovery.ocel.link_analysis.variants import (
    classic as link_analysis,
)


class Parameters(Enum):
    SORTING_COLUMN = "sorting_column"
    INDEX_KEY = "index_key"
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    IN_COLUMN = "in_column"
    OUT_COLUMN = "out_column"
    NODE_COLUMN_SOURCE = "node_column_source"
    NODE_COLUMN_TARGET = "node_column_target"
    EDGE_COLUMN = "edge_column"
    INCLUDE_PERFORMANCE = "include_performance"
    BUSINESS_HOURS = "business_hours"
    BUSINESS_HOUR_SLOTS = "business_hour_slots"
    WORKCALENDAR = "workcalendar"
    TIMESTAMP_DIFF_COLUMN = "timestamp_diff_column"
    EDGE_REFERENCE = "edge_reference"


def build_network_analysis_from_link_analysis(
    merged_df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
) -> Dict[Tuple[str, str], Dict[str, Any]]:
    """
    Builds the network analysis from the results of the link analysis (internal method)

    Parameters
    ------------------
    merged_df
        Dataframe obtained from the link analysis
    parameters
        Parameters of the method, including:
            - Parameters.NODE_COLUMN_SOURCE => the attribute to be used for the node definition of the source event (default: the resource of the log, org:resource)
            - Parameters.NODE_COLUMN_TARGET => the attribute to be used for the node definition of the target event (default: the resource of the log, org:resource)
            - Parameters.EDGE_COLUMN => the attribute to be used for the edge definition (default: the activity of the log, concept:name)
            - Parameters.EDGE_REFERENCE => the event into which the edge attribute should be picked:
                - _out  =>  the source event
                - _in   =>  the target event
            - Parameters.TIMESTAMP_COLUMN => the timestamp column
            - Parameters.TIMESTAMP_DIFF_COLUMN => timestamp diff column
            - Parameters.INCLUDE_PERFORMANCE => considers the performance of the edge
            - Parameters.BUSINESS_HOURS => boolean value that enables the business hours
            - Parameters.BUSINESS_HOURS_SLOTS =>
            work schedule of the company, provided as a list of tuples where each tuple represents one time slot of business
            hours. One slot i.e. one tuple consists of one start and one end time given in seconds since week start, e.g.
            [
                (7 * 60 * 60, 17 * 60 * 60),
                ((24 + 7) * 60 * 60, (24 + 12) * 60 * 60),
                ((24 + 13) * 60 * 60, (24 + 17) * 60 * 60),
            ]
            meaning that business hours are Mondays 07:00 - 17:00 and Tuesdays 07:00 - 12:00 and 13:00 - 17:00

    Returns
    -----------------
    network_analysis
        Edges of the network analysis (first key: edge; second key: type; value: number of occurrences)
    """
    if parameters is None:
        parameters = {}

    node_column_source = exec_utils.get_param_value(
        Parameters.NODE_COLUMN_SOURCE,
        parameters,
        xes_constants.DEFAULT_RESOURCE_KEY,
    )
    node_column_target = exec_utils.get_param_value(
        Parameters.NODE_COLUMN_TARGET,
        parameters,
        xes_constants.DEFAULT_RESOURCE_KEY,
    )
    edge_column = exec_utils.get_param_value(
        Parameters.EDGE_COLUMN, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    edge_reference = exec_utils.get_param_value(
        Parameters.EDGE_REFERENCE, parameters, "_out"
    )
    timestamp_column = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    timestamp_diff_column = exec_utils.get_param_value(
        Parameters.TIMESTAMP_DIFF_COLUMN, parameters, "@@timestamp_diff"
    )

    include_performance = exec_utils.get_param_value(
        Parameters.INCLUDE_PERFORMANCE, parameters, False
    )
    business_hours = exec_utils.get_param_value(
        Parameters.BUSINESS_HOURS, parameters, False
    )
    business_hours_slots = exec_utils.get_param_value(
        Parameters.BUSINESS_HOUR_SLOTS,
        parameters,
        constants.DEFAULT_BUSINESS_HOUR_SLOTS,
    )

    edges = {}

    if business_hours:
        merged_df[timestamp_diff_column] = merged_df.apply(
            lambda x: soj_time_business_hours_diff(
                x[timestamp_column + "_out"],
                x[timestamp_column + "_in"],
                business_hours_slots,
            ),
            axis=1,
        )

    else:
        merged_df[timestamp_diff_column] = pandas_utils.get_total_seconds(
            merged_df[timestamp_column + "_in"]
            - merged_df[timestamp_column + "_out"]
        )

    edges0 = (
        merged_df.dropna(
            subset=[
                node_column_source + "_out",
                node_column_target + "_in",
                edge_column + edge_reference,
            ],
            how="any",
        )
        .groupby(
            [
                node_column_source + "_out",
                node_column_target + "_in",
                edge_column + edge_reference,
            ]
        )[timestamp_diff_column]
        .agg(list)
        .to_dict()
    )

    for e0 in edges0:
        edge = (e0[0], e0[1])
        edge_value = e0[2]
        if edge not in edges:
            edges[edge] = {}
        if edge_value not in edges[edge]:
            if include_performance:
                edges[edge][edge_value] = edges0[e0]
            else:
                edges[edge][edge_value] = len(edges0[e0])

    return edges


def apply(
    dataframe: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
) -> Dict[Tuple[str, str], Dict[str, Any]]:
    """
    Performs the network analysis on the provided dataframe

    Parameters
    -----------------
    dataframe
        Dataframe
    parameters
        Parameters of the method, including:
        - Parameters.SORTING_COLUMN => the column that should be used to sort the log
        - Parameters.IN_COLUMN => the target column of the link (default: the case identifier; events of the same case are linked)
        - Parameters.OUT_COLUMN => the source column of the link (default: the case identifier; events of the same case are linked)
        - Parameters.INDEX_KEY => the name for the index attribute in the log (inserted during the execution)
        - Parameters.NODE_COLUMN_SOURCE => the attribute to be used for the node definition of the source event (default: the resource of the log, org:resource)
        - Parameters.NODE_COLUMN_TARGET => the attribute to be used for the node definition of the target event (default: the resource of the log, org:resource)
        - Parameters.EDGE_COLUMN => the attribute to be used for the edge definition (default: the activity of the log, concept:name)
        - Parameters.EDGE_REFERENCE => the event into which the edge attribute should be picked:
                - _out  =>  the source event
                - _in   =>  the target event
        - Parameters.TIMESTAMP_COLUMN => the timestamp column
        - Parameters.TIMESTAMP_DIFF_COLUMN => timestamp diff column
        - Parameters.INCLUDE_PERFORMANCE => considers the performance of the edge
        - Parameters.BUSINESS_HOURS => boolean value that enables the business hours
        - Parameters.BUSINESS_HOURS_SLOTS =>
        work schedule of the company, provided as a list of tuples where each tuple represents one time slot of business
        hours. One slot i.e. one tuple consists of one start and one end time given in seconds since week start, e.g.
        [
            (7 * 60 * 60, 17 * 60 * 60),
            ((24 + 7) * 60 * 60, (24 + 12) * 60 * 60),
            ((24 + 13) * 60 * 60, (24 + 17) * 60 * 60),
        ]
        meaning that business hours are Mondays 07:00 - 17:00 and Tuesdays 07:00 - 12:00 and 13:00 - 17:00

    Returns
    -----------------
    network_analysis
        Edges of the network analysis (first key: edge; second key: type; value: number of occurrences)
    """
    if parameters is None:
        parameters = {}

    sorting_column = exec_utils.get_param_value(
        Parameters.SORTING_COLUMN,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    index_key = exec_utils.get_param_value(
        Parameters.INDEX_KEY, parameters, constants.DEFAULT_INDEX_KEY
    )
    timestamp_column = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    in_column = exec_utils.get_param_value(
        Parameters.IN_COLUMN, parameters, constants.CASE_CONCEPT_NAME
    )
    out_column = exec_utils.get_param_value(
        Parameters.OUT_COLUMN, parameters, constants.CASE_CONCEPT_NAME
    )
    node_column_source = exec_utils.get_param_value(
        Parameters.NODE_COLUMN_SOURCE,
        parameters,
        xes_constants.DEFAULT_RESOURCE_KEY,
    )
    node_column_target = exec_utils.get_param_value(
        Parameters.NODE_COLUMN_TARGET,
        parameters,
        xes_constants.DEFAULT_RESOURCE_KEY,
    )
    edge_column = exec_utils.get_param_value(
        Parameters.EDGE_COLUMN, parameters, xes_constants.DEFAULT_NAME_KEY
    )

    dataframe = dataframe[
        list(
            {
                timestamp_column,
                in_column,
                out_column,
                node_column_source,
                node_column_target,
                edge_column,
                sorting_column,
            }
        )
    ]

    parameters_la = {
        link_analysis.Parameters.OUT_COLUMN: out_column,
        link_analysis.Parameters.IN_COLUMN: in_column,
        link_analysis.Parameters.INDEX_COLUMN: index_key,
        link_analysis.Parameters.SORTING_COLUMN: sorting_column,
        link_analysis.Parameters.LOOK_FORWARD: True,
        link_analysis.Parameters.KEEP_FIRST_OCCURRENCE: True,
        link_analysis.Parameters.PROPAGATE: False,
    }

    merged_df = link_analysis.apply(dataframe, parameters=parameters_la)

    return build_network_analysis_from_link_analysis(
        merged_df, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils, constants
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
from pm4py.objects.conversion.log import converter as log_converter
import pandas as pd
from copy import copy
from pm4py.util.dt_parsing.variants import strpfromiso
import importlib.util


class Parameters(Enum):
    RETURN_LEGACY_LOG_OBJECT = "return_legacy_log_object"
    RETURN_PL_LAZYFRAME = "return_pl_lazyframe"


def apply(
    log_path: str, parameters: Optional[Dict[Any, Any]] = None
) -> Union[EventLog, pd.DataFrame]:
    if parameters is None:
        parameters = {}

    return_legacy_log_object = exec_utils.get_param_value(
        Parameters.RETURN_LEGACY_LOG_OBJECT, parameters, True
    )
    return_pl_lazyframe = exec_utils.get_param_value(Parameters.RETURN_PL_LAZYFRAME, parameters, False)

    import rustxes

    log = rustxes.import_xes(log_path)
    log = log[0]

    if not return_pl_lazyframe:
        log = log.to_pandas()
        for col in log.columns:
            if "date" in str(log[col].dtype) or "time" in str(log[col].dtype):
                log[col] = strpfromiso.fix_dataframe_column(log[col])
    else:
        import polars as pl

        log = log.lazy()

        schema = log.collect_schema()
        datetime_expressions = []
        for col_name, dtype in schema.items():
            dtype_str = str(dtype).lower()
            if "date" in dtype_str or "time" in dtype_str:
                expr = pl.col(col_name)
                has_timezone = "time_zone=" in dtype_str and "time_zone=none" not in dtype_str
                if constants.ENABLE_DATETIME_COLUMNS_AWARE:
                    expr = (
                        expr.dt.convert_time_zone("UTC")
                        if has_timezone
                        else expr.dt.replace_time_zone("UTC")
                    )
                    datetime_expressions.append(expr.alias(col_name))
                elif has_timezone:
                    datetime_expressions.append(expr.dt.replace_time_zone(None).alias(col_name))
        if datetime_expressions:
            log = log.with_columns(datetime_expressions)

        log = log.collect().lazy()

    if importlib.util.find_spec("cudf") and not return_pl_lazyframe:
        import cudf

        log = cudf.DataFrame(log)

    if return_legacy_log_object:
        this_parameters = copy(parameters)
        this_parameters["stream_postprocessing"] = True

        log = log_converter.apply(
            log,
            variant=log_converter.Variants.TO_EVENT_LOG,
            parameters=this_parameters,
        )

    return log



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.organizational_mining.network_analysis.variants import (
    dataframe,
)
from enum import Enum
from pm4py.util import exec_utils
from typing import Dict, Optional, Any, Tuple, Union
import pandas as pd
from pm4py.objects.log.obj import EventLog, EventStream
from pm4py.objects.conversion.log import converter as log_converter


class Variants(Enum):
    DATAFRAME = dataframe


def apply(
    log: Union[pd.DataFrame, EventLog, EventStream],
    variant=Variants.DATAFRAME,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[Tuple[str, str], Dict[str, Any]]:
    """
    Performs the network analysis on the provided event log

    Parameters
    ----------------
    log
        Event log
    parameters
        Version-specific parameters

    Returns
    ----------------
    network_analysis
        Edges of the network analysis (first key: edge; second key: type; value: number of occurrences)
    """
    return exec_utils.get_variant(variant).apply(
        log_converter.apply(
            log,
            variant=log_converter.Variants.TO_DATA_FRAME,
            parameters=parameters,
        ),
        parameters=parameters,
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from pm4py.objects.conversion.ocel.variants import (
    ocel_to_nx,
    ocel_features_to_nx,
)


class Variants(Enum):
    OCEL_TO_NX = ocel_to_nx
    OCEL_FEATURES_TO_NX = ocel_features_to_nx


def apply(
    ocel: OCEL,
    variant=Variants.OCEL_TO_NX,
    parameters: Optional[Dict[Any, Any]] = None,
):
    """
    Converts an OCEL to another object.

    Parameters
    -------------
    ocel
        Object-centric event log
    variant
        Variant of the algorithm to use, posible values:
        - Variants.OCEL_TO_NX: graph containing event and object IDS and two type of relations (REL=related objects, DF=directly-follows)
        - Variants.OCEL_FEATURES_TO_NX: graph containing different types of interconnection at the object level
    parameters
        Variant-specific parameters
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).apply(ocel, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.ocel.obj import OCEL
from pm4py.algo.discovery.ocel.ocdfg.variants import classic as ocdfg_discovery
from pm4py.algo.discovery.inductive import algorithm as inductive_miner
from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.ocel import constants as ocel_constants
from collections import Counter
from typing import Optional, Dict, Any
from pm4py.objects.dfg.obj import DFG
from pm4py.objects.conversion.process_tree import converter as tree_converter
from pm4py.algo.conformance.tokenreplay import algorithm as token_based_replay
from pm4py.objects.ocel.util import flattening
from copy import copy


class Parameters(Enum):
    EVENT_ACTIVITY = ocel_constants.PARAM_EVENT_ACTIVITY
    OBJECT_TYPE = ocel_constants.PARAM_OBJECT_TYPE
    INDUCTIVE_MINER_VARIANT = "inductive_miner_variant"
    DOUBLE_ARC_THRESHOLD = "double_arc_threshold"
    DIAGNOSTICS_WITH_TBR = "diagnostics_with_token_based_replay"


def apply(
    ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> Dict[str, Any]:
    """
    Discovers an object-centric Petri net (without annotation) from the given object-centric event log,
    using the Inductive Miner as process discovery algorithm.

    Reference paper: van der Aalst, Wil MP, and Alessandro Berti. "Discovering object-centric Petri nets." Fundamenta informaticae 175.1-4 (2020): 1-40.

    Parameters
    -----------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ACTIVITY => the activity attribute to be used
        - Parameters.OBJECT_TYPE => the object type attribute to be used
        - Parameters.DOUBLE_ARC_THRESHOLD => the threshold for the attribution of the "double arc", as
        described in the paper.
        - Parameters.DIAGNOSTICS_WITH_TBR => performs token-based replay and stores the result in the return dict

    Returns
    -----------------
    ocpn
        Object-centric Petri net model, presented as a dictionary of properties:
        - activities: complete set of activities derived from the object-centric event log
        - object_types: complete set of object types derived from the object-centric event log
        - edges: dictionary connecting each object type to a set of directly-followed arcs between activities (expressed as tuples,
                  e.g., (act1, act2)). Every pair of activities is linked to some sets:
                - event_pairs: the tuples of event identifiers where the directly-follows arc occurs
                - total_objects: set of tuples containing two event and one object identifier, uniquely identifying an
                                  occurrence of the arc.
        - activities_indep: dictionary linking each activity, regardless of the object type, to some sets:
            - events: the event identifiers where the activity occurs
            - unique_objects: the object identifiers where the activity occurs
            - total_objects: the tuples of event and object identifiers where the activity occurs.
        - activities_ot: dictionary linking each object type to another dictionary, where the activities are linked to some sets:
            - events: the event identifiers where the activity occurs (with at least one object of the given object type)
            - unique_objects: the object identifiers of the given object type where the activity occurs
            - total_objects: the tuples of event and object identifiers where the activity occurs.
        - start_activities: dictionary linking each object type to another dictionary, where the start activities
                            of the given object type are linked to some sets:
            - events: the event identifiers where the start activity occurs (with at least one object of the given object type)
            - unique_objects: the object identifiers of the given object type where the start activity occurs
            - total_objects: the tuples of event and object identifiers where the start activity occurs.
        - end_activities: dictionary linking each object type to another dictionary, where the end activities
                          of the given object type are linked to some sets:
            - events: the event identifiers where the end activity occurs (with at least one object of the given object type)
            - unique_objects: the object identifiers of the given object type where the end activity occurs
            - total_objects: the tuples of event and object identifiers where the end activity occurs.
        - petri_nets: the accepted Petri nets (Petri net + initial marking + final marking) discovered by the process discovery algorithm
        - double_arcs_on_activity: dictionary linking each object type to another dictionary, where each arc of the Petri net
                                    is linked to a boolean (True if it is a double arc)
        - tbr_results: the results of the token-based replay operation (if required)
    """
    if parameters is None:
        parameters = {}

    double_arc_threshold = exec_utils.get_param_value(
        Parameters.DOUBLE_ARC_THRESHOLD, parameters, 0.8
    )
    inductive_miner_variant = exec_utils.get_param_value(
        Parameters.INDUCTIVE_MINER_VARIANT, parameters, "im"
    )
    diagnostics_with_tbr = exec_utils.get_param_value(
        Parameters.DIAGNOSTICS_WITH_TBR, parameters, False
    )

    ocdfg_parameters = copy(parameters)
    ocdfg_parameters["compute_edges_performance"] = False
    ocpn = ocdfg_discovery.apply(ocel, parameters=ocdfg_parameters)

    petri_nets = {}
    double_arcs_on_activity = {}
    tbr_results = {}

    for ot in ocpn["object_types"]:
        activities_eo = ocpn["activities_ot"]["total_objects"][ot]

        start_activities = {
            x: len(y)
            for x, y in ocpn["start_activities"]["events"][ot].items()
        }
        end_activities = {
            x: len(y) for x, y in ocpn["end_activities"]["events"][ot].items()
        }
        dfg = {}
        if ot in ocpn["edges"]["event_couples"]:
            dfg = {
                x: len(y)
                for x, y in ocpn["edges"]["event_couples"][ot].items()
            }

        is_activity_double = {}
        for act in activities_eo:
            ev_obj_count = Counter([x[0] for x in activities_eo[act]])
            this_single_amount = 0
            for y in ev_obj_count.values():
                if y == 1:
                    this_single_amount += 1
            this_single_amount = this_single_amount / len(ev_obj_count)

            if this_single_amount <= double_arc_threshold:
                is_activity_double[act] = True
            else:
                is_activity_double[act] = False

        double_arcs_on_activity[ot] = is_activity_double

        im_parameters = copy(parameters)
        # disables the fallthroughs, as computing the model on a myriad of different object types
        # could be really expensive
        im_parameters["disable_fallthroughs"] = True
        # for performance reasons, also disable the strict sequence cut (use
        # the normal sequence cut)
        im_parameters["disable_strict_sequence_cut"] = True

        process_tree = None
        flat_log = None

        if inductive_miner_variant == "im" or diagnostics_with_tbr:
            # do the flattening only if it is required
            flat_log = flattening.flatten(ocel, ot, parameters=parameters)

        if inductive_miner_variant == "imd":
            obj = DFG()
            obj._graph = Counter(dfg)
            obj._start_activities = Counter(start_activities)
            obj._end_activities = Counter(end_activities)
            process_tree = inductive_miner.apply(
                obj,
                variant=inductive_miner.Variants.IMd,
                parameters=im_parameters,
            )
        elif inductive_miner_variant == "im":
            process_tree = inductive_miner.apply(
                flat_log, parameters=im_parameters
            )

        petri_net = tree_converter.apply(process_tree, parameters=parameters)

        if diagnostics_with_tbr:
            tbr_parameters = copy(parameters)
            tbr_parameters["enable_pltr_fitness"] = True
            tbr_parameters["show_progress_bar"] = False

            (
                replayed_traces,
                place_fitness_per_trace,
                transition_fitness_per_trace,
                notexisting_activities_in_model,
            ) = token_based_replay.apply(
                flat_log,
                petri_net[0],
                petri_net[1],
                petri_net[2],
                parameters=tbr_parameters,
            )
            place_diagnostics = {
                place: {"m": 0, "r": 0, "c": 0, "p": 0}
                for place in place_fitness_per_trace
            }
            trans_count = {trans: 0 for trans in petri_net[0].transitions}
            # computes the missing, remaining, consumed, and produced tokens
            # per place.
            for place, res in place_fitness_per_trace.items():
                place_diagnostics[place]["m"] += res["m"]
                place_diagnostics[place]["r"] += res["r"]
                place_diagnostics[place]["c"] += res["c"]
                place_diagnostics[place]["p"] += res["p"]

            # counts the number of times a transition has been fired during the
            # replay.
            for trace in replayed_traces:
                for trans in trace["activated_transitions"]:
                    trans_count[trans] += 1

            tbr_results[ot] = (place_diagnostics, trans_count)

        petri_nets[ot] = petri_net

    ocpn["petri_nets"] = petri_nets
    ocpn["double_arcs_on_activity"] = double_arcs_on_activity
    ocpn["tbr_results"] = tbr_results

    return ocpn



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from enum import Enum
from pm4py.util import exec_utils, pandas_utils


class Parameters(Enum):
    PREFIX_OR_SUFFIX = "prefix_or_suffix"
    INDEX_ATTRIBUTE = "index_attribute"
    LEFT_SUFFIX = "left_suffix"
    RIGHT_SUFFIX = "right_suffix"


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    """
    Gets for each event and each one of the related objects,
    the prefix or the suffix.
    E.g. if there is an object "o1" with lifecycle "e1", "e2", "e3",
    then:

    The prefix of "e1" for "o1" is empty.
    The prefix of "e2" for "o1" is ["e1"].
    The prefix of "e3" for "o1" is ["e1", "e2"].

    The suffix of "e1" for "o1" is ["e2", "e3"].
    The suffix of "e2" for "o1" is ["o3"].
    The suffix of "e3" for "o1" is empty.

    Parameters
    --------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm, including:
            - Parameters.PREFIX_OR_SUFFIX => establishes if the prefix or the suffix of the event per object is
                                                needed. Possible values:
                                                    - "prefix"
                                                    - "suffix"
            - Parameters.INDEX_ATTRIBUTE => the index attribute which is artificially inserted in the relations df.
            - Parameters.LEFT_SUFFIX => the suffix for the left-part of the merge.
            - Parameters.RIGHT_SUFFIX => the suffix for the right-part of the merge

    Returns
    --------------
    ret
        A dictionary where the first key is the event, the second key is the object, and the value is a list
        of events which are the object-specific prefix or suffix of the event.
    """
    if parameters is None:
        parameters = {}

    prefix_or_suffix = exec_utils.get_param_value(
        Parameters.PREFIX_OR_SUFFIX, parameters, "prefix"
    )
    index_attribute = exec_utils.get_param_value(
        Parameters.INDEX_ATTRIBUTE, parameters, "@@index"
    )
    left_suffix = exec_utils.get_param_value(
        Parameters.LEFT_SUFFIX, parameters, "_LEFT"
    )
    right_suffix = exec_utils.get_param_value(
        Parameters.RIGHT_SUFFIX, parameters, "_RIGHT"
    )

    relations = ocel.relations.copy()
    relations = pandas_utils.insert_index(
        relations, index_attribute, reset_index=False, copy_dataframe=False
    )
    relations_merged = relations.merge(
        relations,
        left_on=ocel.object_id_column,
        right_on=ocel.object_id_column,
        suffixes=(left_suffix, right_suffix),
    )
    if prefix_or_suffix == "prefix":
        relations_merged = relations_merged[
            relations_merged[index_attribute + left_suffix]
            > relations_merged[index_attribute + right_suffix]
        ]
    else:
        relations_merged = relations_merged[
            relations_merged[index_attribute + left_suffix]
            < relations_merged[index_attribute + right_suffix]
        ]
    relations_merged = relations_merged[
        [
            ocel.event_id_column + left_suffix,
            ocel.event_id_column + right_suffix,
            ocel.object_id_column,
        ]
    ]
    relations_merged = relations_merged[
        relations_merged[ocel.event_id_column + left_suffix]
        != relations_merged[ocel.event_id_column + right_suffix]
    ]
    relations_merged = relations_merged.to_dict("records")

    ret = {}
    for el in relations_merged:
        e1 = el[ocel.event_id_column + left_suffix]
        e2 = el[ocel.event_id_column + right_suffix]
        obj = el[ocel.object_id_column]
        if e1 not in ret:
            ret[e1] = {}
        if obj not in ret[e1]:
            ret[e1][obj] = []
        if e2 not in ret[e1][obj]:
            ret[e1][obj].append(e2)

    return ret



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
import warnings
from pm4py.util import pandas_utils


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None) -> OCEL:
    """
    Forces the consistency of the OCEL, ensuring that the event/object identifier,
    event/object type are of type string and non-empty.

    Parameters
    --------------
    ocel
        OCEL
    parameters
        Possible parameters of the method

    Returns
    --------------
    ocel
        Consistent OCEL
    """
    if parameters is None:
        parameters = {}

    # Store frequently accessed column names locally to reduce attribute lookups
    event_id_col = ocel.event_id_column
    object_id_col = ocel.object_id_column
    event_activity = ocel.event_activity
    object_type_col = ocel.object_type_column

    # Define fields to process for each dataframe
    fields = {
        "events": [event_id_col, event_activity],
        "objects": [object_id_col, object_type_col],
        "relations": [event_id_col, object_id_col, event_activity, object_type_col],
        "o2o": [object_id_col, object_id_col + "_2"],
        "e2e": [event_id_col, event_id_col + "_2"],
        "object_changes": [object_id_col],
    }

    # Process each dataframe
    for tab, columns in fields.items():
        # Skip processing if attribute doesn't exist
        if not hasattr(ocel, tab):
            continue

        # Get dataframe
        df = getattr(ocel, tab)

        # Skip empty dataframes
        if df.empty:
            continue

        # Filter to only columns that exist in this dataframe
        valid_columns = [col for col in columns if col in df.columns]
        if not valid_columns:
            continue

        # Check for NA values - only create mask if needed
        has_na = df[valid_columns].isna().any().any()
        if has_na:
            # Create mask for rows without NA values
            valid_rows = ~df[valid_columns].isna().any(axis=1)
            df = df.loc[valid_rows]

        # Convert columns to string type
        for col in valid_columns:
            df[col] = df[col].astype(str)

        # Efficiently filter out empty strings
        # Create a single mask for all columns and apply once
        valid_rows = pandas_utils.DATAFRAME.Series(True, index=df.index)
        for col in valid_columns:
            valid_rows &= (df[col].str.len() > 0)

        # Only filter if we found empty strings
        if not valid_rows.all():
            df = df.loc[valid_rows]

        # Update OCEL attribute
        setattr(ocel, tab, df)

    # Check uniqueness efficiently
    events_df = ocel.events
    objects_df = ocel.objects

    # Only check if there are rows to check
    if len(events_df) > 0:
        num_ev_ids = events_df[event_id_col].nunique()
        if num_ev_ids < len(events_df):
            warnings.warn("The event identifiers in the OCEL are not unique!")

    if len(objects_df) > 0:
        num_obj_ids = objects_df[object_id_col].nunique()
        if num_obj_ids < len(objects_df):
            warnings.warn("The object identifiers in the OCEL are not unique!")

    return ocel



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME
from pm4py.util.xes_constants import DEFAULT_NAME_KEY
from pm4py.util.constants import PARAMETER_CONSTANT_ACTIVITY_KEY
from pm4py.util.constants import PARAMETER_CONSTANT_ATTRIBUTE_KEY
from pm4py.util.constants import PARAMETER_CONSTANT_CASEID_KEY
from enum import Enum
from pm4py.util import exec_utils
from copy import copy
from typing import Optional, Dict, Any, Union, List
import polars as pl


class Parameters(Enum):
    ATTRIBUTE_KEY = PARAMETER_CONSTANT_ATTRIBUTE_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    DECREASING_FACTOR = "decreasingFactor"
    POSITIVE = "positive"
    STREAM_FILTER_KEY1 = "stream_filter_key1"
    STREAM_FILTER_VALUE1 = "stream_filter_value1"
    STREAM_FILTER_KEY2 = "stream_filter_key2"
    STREAM_FILTER_VALUE2 = "stream_filter_value2"
    KEEP_ONCE_PER_CASE = "keep_once_per_case"


def get_attribute_values(
    df: pl.LazyFrame,
    attribute_key: str,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[Any, int]:
    """
    Return list of attribute values contained in the specified column

    Parameters
    -----------
    df
        Polars LazyFrame
    attribute_key
        Attribute for which we want to known the values and the count
    parameters
        Possible parameters of the algorithm

    Returns
    -----------
    attributes_values_dict
        Attributes in the specified column, along with their count
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    keep_once_per_case = exec_utils.get_param_value(
        Parameters.KEEP_ONCE_PER_CASE, parameters, False
    )

    if keep_once_per_case:
        # Get unique values per case
        result = (
            df.group_by([case_id_glue, attribute_key])
            .first()
            .group_by(attribute_key)
            .count()
            .collect()
        )
    else:
        # Count all occurrences
        result = (
            df.group_by(attribute_key)
            .count()
            .collect()
        )
    
    # Convert to dictionary
    attributes_values_dict = {}
    for row in result.iter_rows():
        attributes_values_dict[row[0]] = row[1]
    
    return attributes_values_dict


def apply_numeric_events(
    df: pl.LazyFrame,
    int1: float,
    int2: float,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Apply a filter on events (numerical filter)

    Parameters
    --------------
    df
        LazyFrame
    int1
        Lower bound of the interval
    int2
        Upper bound of the interval
    parameters
        Possible parameters of the algorithm:
            Parameters.ATTRIBUTE_KEY => indicates which attribute to filter
            positive => keep or remove events?

    Returns
    --------------
    filtered_df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    if positive:
        ret = df.filter(
            (pl.col(attribute_key) >= int1) & (pl.col(attribute_key) <= int2)
        )
    else:
        ret = df.filter(
            (pl.col(attribute_key) < int1) | (pl.col(attribute_key) > int2)
        )

    return ret


def apply_numeric(
    df: pl.LazyFrame,
    int1: float,
    int2: float,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Filter LazyFrame on attribute values (filter cases)

    Parameters
    --------------
    df
        LazyFrame
    int1
        Lower bound of the interval
    int2
        Upper bound of the interval
    parameters
        Possible parameters of the algorithm:
            Parameters.ATTRIBUTE_KEY => indicates which attribute to filter
            Parameters.POSITIVE => keep or remove traces with such events?

    Returns
    --------------
    filtered_df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    # stream_filter_key is helpful to filter on cases containing an event with an attribute
    # in the specified value set, but such events shall have an activity in
    # particular.
    stream_filter_key1 = exec_utils.get_param_value(
        Parameters.STREAM_FILTER_KEY1, parameters, None
    )
    stream_filter_value1 = exec_utils.get_param_value(
        Parameters.STREAM_FILTER_VALUE1, parameters, None
    )
    stream_filter_key2 = exec_utils.get_param_value(
        Parameters.STREAM_FILTER_KEY2, parameters, None
    )
    stream_filter_value2 = exec_utils.get_param_value(
        Parameters.STREAM_FILTER_VALUE2, parameters, None
    )

    # Filter events by numeric range
    filtered_df_by_ev = df.filter(
        (pl.col(attribute_key) >= int1) & (pl.col(attribute_key) <= int2)
    )
    
    if stream_filter_key1 is not None:
        filtered_df_by_ev = filtered_df_by_ev.filter(
            pl.col(stream_filter_key1) == stream_filter_value1
        )
    if stream_filter_key2 is not None:
        filtered_df_by_ev = filtered_df_by_ev.filter(
            pl.col(stream_filter_key2) == stream_filter_value2
        )

    # Get case IDs that match the criteria
    matching_cases = filtered_df_by_ev.select(case_id_glue).unique()
    
    if positive:
        # Keep cases that match
        ret = df.join(matching_cases, on=case_id_glue, how="inner")
    else:
        # Keep cases that don't match
        ret = df.join(matching_cases, on=case_id_glue, how="anti")

    return ret


def apply_events(
    df: pl.LazyFrame,
    values: List[str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Filter LazyFrame on attribute values (filter events)

    Parameters
    ----------
    df
        LazyFrame
    values
        Values to filter on
    parameters
        Possible parameters of the algorithm, including:
            Parameters.ATTRIBUTE_KEY -> Attribute we want to filter
            Parameters.POSITIVE -> Specifies if the filter should be applied including traces (positive=True) or
            excluding traces (positive=False)
    Returns
    ----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    if positive:
        ret = df.filter(pl.col(attribute_key).is_in(values))
    else:
        ret = df.filter(~pl.col(attribute_key).is_in(values))

    return ret


def apply(
    df: pl.LazyFrame,
    values: List[str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Filter LazyFrame on attribute values (filter traces)

    Parameters
    ----------
    df
        LazyFrame
    values
        Values to filter on
    parameters
        Possible parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Case ID column in the LazyFrame
            Parameters.ATTRIBUTE_KEY -> Attribute we want to filter
            Parameters.POSITIVE -> Specifies if the filter should be applied including traces (positive=True) or
            excluding traces (positive=False)
    Returns
    ----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    return filter_df_on_attribute_values(
        df,
        values,
        case_id_glue=case_id_glue,
        attribute_key=attribute_key,
        positive=positive,
    )


def filter_df_on_attribute_values(
    df: pl.LazyFrame,
    values: List[str],
    case_id_glue: str = "case:concept:name",
    attribute_key: str = "concept:name",
    positive: bool = True,
) -> pl.LazyFrame:
    """
    Filter LazyFrame on attribute values

    Parameters
    ----------
    df
        LazyFrame
    values
        Values to filter on
    case_id_glue
        Case ID column in the LazyFrame
    attribute_key
        Attribute we want to filter
    positive
        Specifies if the filtered should be applied including traces (positive=True) or excluding traces
        (positive=False)

    Returns
    ----------
    df
        Filtered LazyFrame
    """
    if values is None:
        values = []
    
    # Get events matching the attribute values
    filtered_df_by_ev = df.filter(pl.col(attribute_key).is_in(values))
    
    # Get unique case IDs from filtered events
    matching_cases = filtered_df_by_ev.select(case_id_glue).unique()
    
    if positive:
        # Keep cases that have matching events
        ret = df.join(matching_cases, on=case_id_glue, how="inner")
    else:
        # Keep cases that don't have matching events
        ret = df.join(matching_cases, on=case_id_glue, how="anti")

    return ret


def filter_df_keeping_activ_exc_thresh(
    df: pl.LazyFrame,
    thresh: int,
    act_count0: Optional[Dict[str, int]] = None,
    activity_key: str = "concept:name",
    most_common_variant: Optional[List[str]] = None,
) -> pl.LazyFrame:
    """
    Filter a LazyFrame keeping activities exceeding the threshold

    Parameters
    ------------
    df
        Polars LazyFrame
    thresh
        Threshold to use to cut activities
    act_count0
        (If provided) Dictionary that associates each activity with its count
    activity_key
        Column in which the activity is present
    most_common_variant
        (If provided) List of activities in most common variant

    Returns
    ------------
    df
        Filtered LazyFrame
    """
    if most_common_variant is None:
        most_common_variant = []

    if act_count0 is None:
        act_count0 = get_attribute_values(df, activity_key)
    
    act_count = [
        k
        for k, v in act_count0.items()
        if v >= thresh or k in most_common_variant
    ]
    
    if len(act_count) < len(act_count0):
        ret = df.filter(pl.col(activity_key).is_in(act_count))
    else:
        ret = df

    return ret


def filter_df_keeping_spno_activities(
    df: pl.LazyFrame,
    activity_key: str = "concept:name",
    max_no_activities: int = 25,
) -> pl.LazyFrame:
    """
    Filter a LazyFrame on the specified number of attributes

    Parameters
    -----------
    df
        LazyFrame
    activity_key
        Activity key in LazyFrame (must be specified if different from concept:name)
    max_no_activities
        Maximum allowed number of attributes

    Returns
    ------------
    df
        Filtered LazyFrame
    """
    # Get activity counts
    activity_counts = (
        df.group_by(activity_key)
        .count()
        .sort("count", descending=True)
        .limit(max_no_activities)
        .collect()
    )
    
    # Extract activity names to keep
    activity_to_keep = activity_counts[activity_key].to_list()
    
    # Get total unique activities in original dataframe
    total_activities = df.select(pl.col(activity_key).n_unique()).collect()[0, 0]
    
    if len(activity_to_keep) < total_activities:
        ret = df.filter(pl.col(activity_key).is_in(activity_to_keep))
    else:
        ret = df

    return ret


def filter_df_relative_occurrence_event_attribute(
    df: pl.LazyFrame,
    min_relative_stake: float,
    parameters: Optional[Dict[Any, Any]] = None,
) -> pl.LazyFrame:
    """
    Filters the event log keeping only the events having an attribute value which occurs:
    - in at least the specified (min_relative_stake) percentage of events, when Parameters.KEEP_ONCE_PER_CASE = False
    - in at least the specified (min_relative_stake) percentage of cases, when Parameters.KEEP_ONCE_PER_CASE = True

    Parameters
    -------------------
    df
        Polars LazyFrame
    min_relative_stake
        Minimum percentage of cases (expressed as a number between 0 and 1) in which the attribute should occur.
    parameters
        Parameters of the algorithm, including:
        - Parameters.ATTRIBUTE_KEY => the attribute to use (default: concept:name)
        - Parameters.KEEP_ONCE_PER_CASE => decides the level of the filter to apply
        (if the filter should be applied on the cases, set it to True).

    Returns
    ------------------
    filtered_df
        Filtered Polars LazyFrame
    """
    if parameters is None:
        parameters = {}

    attribute_key = exec_utils.get_param_value(
        PARAMETER_CONSTANT_ATTRIBUTE_KEY, parameters, DEFAULT_NAME_KEY
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    keep_once_per_case = exec_utils.get_param_value(
        Parameters.KEEP_ONCE_PER_CASE, parameters, True
    )

    parameters_cp = copy(parameters)

    activities_occurrences = get_attribute_values(
        df, attribute_key, parameters=parameters_cp
    )

    if keep_once_per_case:
        # filter on cases
        n_cases = df.select(pl.col(case_id_key).n_unique()).collect()[0, 0]
        filtered_attributes = set(
            x
            for x, y in activities_occurrences.items()
            if y >= min_relative_stake * n_cases
        )
    else:
        # filter on events
        n_events = df.select(pl.count()).collect()[0, 0]
        filtered_attributes = set(
            x
            for x, y in activities_occurrences.items()
            if y >= min_relative_stake * n_events
        )

    return apply_events(df, list(filtered_attributes), parameters=parameters)


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd
from typing import Optional, Dict, Any, Union
from pm4py.objects.conversion.log import converter as log_converter
from enum import Enum
from pm4py.util import exec_utils, constants


class Parameters(Enum):
    MAX_LEN = "max_len"
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY


def apply(
    log_obj: Union[EventLog, EventStream, pd.DataFrame],
    parameters: Optional[Dict[Any, Any]] = None,
) -> str:
    if parameters is None:
        parameters = {}

    log_obj = log_converter.apply(
        log_obj,
        variant=log_converter.Variants.TO_DATA_FRAME,
        parameters=parameters,
    )
    max_len = exec_utils.get_param_value(
        Parameters.MAX_LEN, parameters, constants.OPENAI_MAX_LEN
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )

    log_obj = log_obj[[x for x in log_obj.columns if x != case_id_key]]
    cols_dtypes = {x: str(log_obj[x].dtype) for x in log_obj.columns}

    num_nans = {}
    lst_values = []

    for x, y in cols_dtypes.items():
        num_nans[x] = log_obj[x].isna().sum()
        if "object" in y or "string" in y or "bool" in y:
            values = log_obj[x].value_counts().to_dict()
            if len(values) > 1:
                for v, co in values.items():
                    lst_values.append((x, v, co))

    lst_values = sorted(
        lst_values, key=lambda x: (x[2], x[0], x[1]), reverse=True
    )
    ret = {}
    curr_len = 0

    for x, y in cols_dtypes.items():
        if "float" in y or "double" in y or "date" in y:
            quantiles = " quantiles: " + str(
                log_obj[x].quantile([0.0, 0.25, 0.5, 0.75, 1.0]).to_dict()
            )
            nans = " empty: " + str(num_nans[x]) + " "
            curr_len += len(x) + len(nans) + len(quantiles) + 1

            ret[x] = nans + quantiles

    for el in lst_values:
        if curr_len >= max_len:
            break

        if not el[0] in ret:
            stru = " empty: " + str(num_nans[el[0]]) + " values:"
            ret[el[0]] = stru
            curr_len += len(el[0]) + len(stru) + 9

        stru = " (" + str(el[1]) + "; freq. " + str(el[2]) + ")"
        ret[el[0]] += stru
        curr_len += len(stru)

    keys = sorted(list(ret))
    ret = [k + " " + ret[k] for k in keys]
    return "\n".join(ret)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import tempfile
from copy import deepcopy
from enum import Enum

from graphviz import Graph

from pm4py.objects.process_tree.utils import generic
from pm4py.util import exec_utils, constants
from typing import Optional, Dict, Any, Union
from pm4py.objects.process_tree.obj import ProcessTree
import graphviz


class Parameters(Enum):
    FORMAT = "format"
    COLOR_MAP = "color_map"
    ENABLE_DEEPCOPY = "enable_deepcopy"
    FONT_SIZE = "font_size"
    BGCOLOR = "bgcolor"
    RANKDIR = "rankdir"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


# maps the operators to the ProM strings
operators_mapping = {
    "->": "seq",
    "X": "xor",
    "+": "and",
    "*": "xor loop",
    "O": "or",
    "<>": "interleaving",
}


def get_color(node, color_map):
    """
    Gets a color for a node from the color map

    Parameters
    --------------
    node
        Node
    color_map
        Color map
    """
    if node in color_map:
        return color_map[node]
    return "black"


def repr_tree_2(tree, viz, color_map, parameters):
    font_size = exec_utils.get_param_value(
        Parameters.FONT_SIZE, parameters, 15
    )
    font_size = str(font_size)

    this_node_id = str(id(tree))

    if tree.operator is None:
        if tree.label is None:
            viz.node(
                this_node_id,
                "tau",
                style="filled",
                fillcolor="black",
                shape="point",
                width="0.075",
                fontsize=font_size,
            )
        else:
            node_color = get_color(tree, color_map)
            viz.node(
                this_node_id,
                str(tree.label),
                color=node_color,
                fontcolor=node_color,
                fontsize=font_size,
            )
    else:
        node_color = get_color(tree, color_map)
        viz.node(
            this_node_id,
            operators_mapping[str(tree.operator)],
            color=node_color,
            fontcolor=node_color,
            fontsize=font_size,
        )

        for child in tree.children:
            repr_tree_2(child, viz, color_map, parameters)

    if tree.parent is not None:
        viz.edge(str(id(tree.parent)), this_node_id, dirType="none")


def apply(
    tree: ProcessTree,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> graphviz.Graph:
    """
    Obtain a Process Tree representation through GraphViz

    Parameters
    -----------
    tree
        Process tree
    parameters
        Possible parameters of the algorithm

    Returns
    -----------
    gviz
        GraphViz object
    """
    if parameters is None:
        parameters = {}

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    bgcolor = exec_utils.get_param_value(
        Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR
    )
    rankdir = exec_utils.get_param_value(
        Parameters.RANKDIR, parameters, constants.DEFAULT_RANKDIR_GVIZ
    )
    font_size = exec_utils.get_param_value(
        Parameters.FONT_SIZE, parameters, 15
    )
    font_size = str(font_size)
    enable_graph_title = exec_utils.get_param_value(
        Parameters.ENABLE_GRAPH_TITLE,
        parameters,
        constants.DEFAULT_ENABLE_GRAPH_TITLES,
    )
    graph_title = exec_utils.get_param_value(
        Parameters.GRAPH_TITLE, parameters, "Process Tree"
    )

    viz = Graph(
        "pt",
        filename=filename.name,
        engine="dot",
        graph_attr={"bgcolor": bgcolor, "rankdir": rankdir},
    )
    viz.attr("node", shape="ellipse", fixedsize="false")

    if enable_graph_title:
        viz.attr(
            label='<<FONT POINT-SIZE="'
            + str(2 * int(font_size))
            + '">'
            + graph_title
            + "</FONT>>",
            labelloc="top",
        )

    image_format = exec_utils.get_param_value(
        Parameters.FORMAT, parameters, "png"
    )
    color_map = exec_utils.get_param_value(
        Parameters.COLOR_MAP, parameters, {}
    )

    enable_deepcopy = exec_utils.get_param_value(
        Parameters.ENABLE_DEEPCOPY, parameters, True
    )

    if enable_deepcopy:
        # since the process tree object needs to be sorted in the visualization, make a deepcopy of it before
        # proceeding
        tree = deepcopy(tree)
        generic.tree_sort(tree)

    repr_tree_2(tree, viz, color_map, parameters)

    viz.attr(overlap="false")
    viz.attr(splines="false")
    viz.format = image_format.replace("html", "plain-ext")

    return viz



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from copy import deepcopy
from enum import Enum
from typing import List, Dict, Any, Optional, Union

from pm4py.objects.log.obj import EventLog
from pm4py.objects.petri_net.utils import align_utils
from pm4py.util import exec_utils


class Parameters(Enum):
    TARGET_ATTRIBUTE = "target_attribute"
    ENABLE_DEEPCOPY = "enable_deepcopy"


def apply(
    log: EventLog,
    aligned_traces: List[Dict[str, Any]],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> EventLog:
    """
    Enriches a log with the results of the alignment against a model,
    obtained with the parameter 'ret_tuple_as_trans_desc' set to True
    (so the identifiers of the transitions of the model are known).
    In particular, the events that are not move-on-log are enriched with
    the identifier of the corresponding element of the model.

    Parameters
    ----------------
    log
        Event log
    aligned_traces
        Result of the alignments, done with the parameter 'ret_tuple_as_trans_Desc_ set to True.
    parameters
        Parameters of the algorithm:
        - Parameters.TARGET_ATTRIBUTE: attribute that should be used for the enrichment
        - Parameters.ENABLE_DEEPCOPY: deepcopy the event log to not enrich the original log.

    Returns
    ----------------
    enriched_log
        Log enriched with an additional attribute (the identifier of the corresponding element of the model)
    """
    if parameters is None:
        parameters = {}

    target_attribute = exec_utils.get_param_value(
        Parameters.TARGET_ATTRIBUTE, parameters, "@@transition_id"
    )
    enable_deepcopy = exec_utils.get_param_value(
        Parameters.ENABLE_DEEPCOPY, parameters, False
    )

    if enable_deepcopy:
        log = deepcopy(log)

    for i in range(len(aligned_traces)):
        z = 0
        for j in range(len(aligned_traces[i]["alignment"])):
            id_piece = aligned_traces[i]["alignment"][j][0]
            if id_piece[0] != align_utils.SKIP:
                if id_piece[1] != align_utils.SKIP:
                    log[i][z][target_attribute] = id_piece[1]
                z = z + 1

    return log



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.conformance.tokenreplay.variants import token_replay
from pm4py.statistics.variants.log import get as variants_module
from pm4py.objects.petri_net.obj import PetriNet
from pm4py.objects.random_variables.random_variable import RandomVariable
from pm4py.objects.petri_net.utils import performance_map
from pm4py.util import exec_utils, xes_constants
from pm4py.algo.conformance.tokenreplay import algorithm as executor

from enum import Enum
from pm4py.util import constants
from copy import copy


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    TOKEN_REPLAY_VARIANT = "token_replay_variant"
    PARAM_NUM_SIMULATIONS = "num_simulations"
    PARAM_FORCE_DISTRIBUTION = "force_distribution"
    PARAM_ENABLE_DIAGNOSTICS = "enable_diagnostics"
    PARAM_DIAGN_INTERVAL = "diagn_interval"
    PARAM_CASE_ARRIVAL_RATIO = "case_arrival_ratio"
    PARAM_PROVIDED_SMAP = "provided_stochastic_map"
    PARAM_MAP_RESOURCES_PER_PLACE = "map_resources_per_place"
    PARAM_DEFAULT_NUM_RESOURCES_PER_PLACE = "default_num_resources_per_place"
    PARAM_SMALL_SCALE_FACTOR = "small_scale_factor"
    PARAM_MAX_THREAD_EXECUTION_TIME = "max_thread_exec_time"


def get_map_from_log_and_net(
    log,
    net,
    initial_marking,
    final_marking,
    force_distribution=None,
    parameters=None,
):
    """
    Get transition stochastic distribution map given the log and the Petri net

    Parameters
    -----------
    log
        Event log
    net
        Petri net
    initial_marking
        Initial marking of the Petri net
    final_marking
        Final marking of the Petri net
    force_distribution
        If provided, distribution to force usage (e.g. EXPONENTIAL)
    parameters
        Parameters of the algorithm, including:
            Parameters.ACTIVITY_KEY -> activity name
            Parameters.TIMESTAMP_KEY -> timestamp key

    Returns
    -----------
    stochastic_map
        Map that to each transition associates a random variable
    """
    stochastic_map = {}

    if parameters is None:
        parameters = {}

    token_replay_variant = exec_utils.get_param_value(
        Parameters.TOKEN_REPLAY_VARIANT,
        parameters,
        executor.Variants.TOKEN_REPLAY,
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )

    parameters_variants = {
        constants.PARAMETER_CONSTANT_ACTIVITY_KEY: activity_key
    }
    variants_idx = variants_module.get_variants_from_log_trace_idx(
        log, parameters=parameters_variants
    )
    variants = variants_module.convert_variants_trace_idx_to_trace_obj(
        log, variants_idx
    )

    parameters_tr = copy(parameters)
    parameters_tr[token_replay.Parameters.ACTIVITY_KEY] = activity_key
    parameters_tr[token_replay.Parameters.VARIANTS] = variants

    parameters_ses = copy(parameters)

    # do the replay
    aligned_traces = executor.apply(
        log,
        net,
        initial_marking,
        final_marking,
        variant=token_replay_variant,
        parameters=parameters_tr,
    )

    element_statistics = performance_map.single_element_statistics(
        log,
        net,
        initial_marking,
        aligned_traces,
        variants_idx,
        activity_key=activity_key,
        timestamp_key=timestamp_key,
        parameters=parameters_ses,
    )

    for el in element_statistics:
        if (
            type(el) is PetriNet.Transition
            and "performance" in element_statistics[el]
        ):
            values = element_statistics[el]["performance"]

            rand = RandomVariable()
            rand.calculate_parameters(
                values, force_distribution=force_distribution
            )

            no_of_times_enabled = element_statistics[el]["no_of_times_enabled"]
            no_of_times_activated = element_statistics[el][
                "no_of_times_activated"
            ]

            if no_of_times_enabled > 0:
                rand.set_weight(
                    float(no_of_times_activated) / float(no_of_times_enabled)
                )
            else:
                rand.set_weight(0.0)

            stochastic_map[el] = rand

    return stochastic_map



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
"""Polars-based classic process cube variant."""

from bisect import bisect_right
import math
from enum import Enum
from typing import Optional, Dict, Any, Tuple, List

import numpy as np
import polars as pl

from pm4py.util import constants, exec_utils

CASE_ID_COL = constants.CASE_CONCEPT_NAME

# Accepted aggregate functions for Polars pivot
_AGG_FN_ALIASES = {
    "avg": "mean",
    "average": "mean",
    "mean": "mean",
    "sum": "sum",
    "min": "min",
    "max": "max",
    "count": "count",
    "len": "count",
    "median": "median",
    "first": "first",
    "last": "last",
}


class Parameters(Enum):
    MAX_DIVISIONS_X = "max_divisions_x"
    MAX_DIVISIONS_Y = "max_divisions_y"
    AGGREGATION_FUNCTION = "aggregation_function"
    X_BINS = "x_bins"
    Y_BINS = "y_bins"


_NUMERIC_DTYPES = {
    pl.Int8,
    pl.Int16,
    pl.Int32,
    pl.Int64,
    pl.UInt8,
    pl.UInt16,
    pl.UInt32,
    pl.UInt64,
    pl.Float32,
    pl.Float64,
}


def _ensure_polars_df(feature_table: pl.LazyFrame | pl.DataFrame) -> pl.DataFrame:
    if isinstance(feature_table, pl.LazyFrame):
        return feature_table.collect()
    if isinstance(feature_table, pl.DataFrame):
        return feature_table.clone()
    raise TypeError(
        "feature_table must be a Polars LazyFrame or DataFrame"
    )


def _is_numeric_dtype(dtype: pl.DataType) -> bool:
    if dtype in _NUMERIC_DTYPES:
        return True
    dtype_str = str(dtype).lower()
    return any(token in dtype_str for token in ("int", "uint", "float"))


def _prepare_bins(
    series: pl.Series,
    manual_bins: Optional[List[float]],
    max_divisions: int,
) -> List[float]:
    if manual_bins:
        bins = sorted(set(float(b) for b in manual_bins))
        if len(bins) >= 2:
            return bins

    values = series.drop_nulls()
    if values.is_empty():
        return []

    min_val = float(values.min())
    max_val = float(values.max())
    if math.isfinite(min_val) and math.isfinite(max_val):
        if math.isclose(min_val, max_val):
            return [min_val - 0.5, max_val + 0.5]

        max_divisions = max(int(max_divisions), 1)
        bins_array = np.linspace(min_val, max_val, max_divisions + 1)
        bins = sorted(set(float(b) for b in bins_array))
        if len(bins) >= 2:
            return bins
        return [min_val - 0.5, max_val + 0.5]

    return []


def _bin_labels(bins: List[float]) -> List[str]:
    if len(bins) < 2:
        return []
    return [f"[{bins[i]}, {bins[i + 1]}]" for i in range(len(bins) - 1)]


def _assign_bins(series: pl.Series, bins: List[float], name: str) -> pl.Series:
    length = len(series)
    if len(bins) < 2:
        return pl.Series(name, [None] * length)

    labels = _bin_labels(bins)
    last_edge = bins[-1]

    assigned: List[Optional[str]] = []
    for value in series.to_list():
        if value is None:
            assigned.append(None)
            continue
        try:
            numeric_value = float(value)
        except (TypeError, ValueError):
            assigned.append(None)
            continue
        if not math.isfinite(numeric_value):
            assigned.append(None)
            continue

        idx = bisect_right(bins, numeric_value) - 1
        if idx == len(labels) and math.isclose(numeric_value, last_edge, rel_tol=1e-9, abs_tol=1e-9):
            idx -= 1
        if idx < 0 or idx >= len(labels):
            assigned.append(None)
        else:
            assigned.append(labels[idx])

    return pl.Series(name, assigned)


def _normalize_agg_fn(agg_fn: Any) -> str:
    if isinstance(agg_fn, str):
        key = agg_fn.lower()
        return _AGG_FN_ALIASES.get(key, key)
    return "mean"


def _aggregation_expression(column: str, agg_fn: str) -> pl.Expr:
    if agg_fn == "sum":
        return pl.col(column).sum()
    if agg_fn == "min":
        return pl.col(column).min()
    if agg_fn == "max":
        return pl.col(column).max()
    if agg_fn == "median":
        return pl.col(column).median()
    if agg_fn == "first":
        return pl.col(column).first()
    if agg_fn == "last":
        return pl.col(column).last()
    if agg_fn == "count":
        return pl.count()
    # default mean
    return pl.col(column).mean()


def _prefix_columns(df: pl.DataFrame, prefix: str) -> List[str]:
    return [col for col in df.columns if col.startswith(prefix+"_")]


def _numeric_numeric_case(
    df: pl.DataFrame,
    x_col: str,
    y_col: str,
    agg_col: str,
    x_bins: List[float],
    y_bins: List[float],
) -> pl.DataFrame:
    x_bin_series = _assign_bins(df[x_col], x_bins, "x_bin")
    y_bin_series = _assign_bins(df[y_col], y_bins, "y_bin")

    df = df.with_columns([x_bin_series, y_bin_series])

    temp_df = (
        df.filter(pl.col("x_bin").is_not_null() & pl.col("y_bin").is_not_null())
        .select([pl.col(CASE_ID_COL), pl.col("x_bin"), pl.col("y_bin"), pl.col(agg_col)])
    )
    return temp_df


def _numeric_prefix_case(
    df: pl.DataFrame,
    x_col: str,
    agg_col: str,
    bins: List[float],
    prefix_cols: List[str],
    prefix_label: str,
) -> pl.DataFrame:
    x_bin_series = _assign_bins(df[x_col], bins, "x_bin")
    df = df.with_columns(x_bin_series)

    if not prefix_cols:
        return pl.DataFrame({CASE_ID_COL: [], "x_bin": [], "y_bin": [], agg_col: []})

    melted = (
        df.select([CASE_ID_COL, agg_col, "x_bin", *prefix_cols])
        .melt(
            id_vars=[CASE_ID_COL, agg_col, "x_bin"],
            value_vars=prefix_cols,
            variable_name=prefix_label,
            value_name="__value",
        )
        .filter(pl.col("x_bin").is_not_null() & (pl.col("__value").fill_null(0) >= 1))
        .select([
            pl.col(CASE_ID_COL),
            pl.col("x_bin"),
            pl.col(prefix_label).alias("y_bin"),
            pl.col(agg_col),
        ])
    )
    return melted


def _prefix_numeric_case(
    df: pl.DataFrame,
    y_col: str,
    agg_col: str,
    bins: List[float],
    prefix_cols: List[str],
    prefix_label: str,
) -> pl.DataFrame:
    y_bin_series = _assign_bins(df[y_col], bins, "y_bin")
    df = df.with_columns(y_bin_series)

    if not prefix_cols:
        return pl.DataFrame({CASE_ID_COL: [], "x_bin": [], "y_bin": [], agg_col: []})

    melted = (
        df.select([CASE_ID_COL, agg_col, "y_bin", *prefix_cols])
        .melt(
            id_vars=[CASE_ID_COL, agg_col, "y_bin"],
            value_vars=prefix_cols,
            variable_name=prefix_label,
            value_name="__value",
        )
        .filter(pl.col("y_bin").is_not_null() & (pl.col("__value").fill_null(0) >= 1))
        .select([
            pl.col(CASE_ID_COL),
            pl.col(prefix_label).alias("x_bin"),
            pl.col("y_bin"),
            pl.col(agg_col),
        ])
    )
    return melted


def _prefix_prefix_case(
    df: pl.DataFrame,
    agg_col: str,
    x_prefix_cols: List[str],
    y_prefix_cols: List[str],
) -> pl.DataFrame:
    if not x_prefix_cols or not y_prefix_cols:
        return pl.DataFrame({CASE_ID_COL: [], "x_bin": [], "y_bin": [], agg_col: []})

    x_melt = (
        df.select([CASE_ID_COL, agg_col, *x_prefix_cols])
        .melt(
            id_vars=[CASE_ID_COL, agg_col],
            value_vars=x_prefix_cols,
            variable_name="x_bin",
            value_name="__x_value",
        )
        .filter(pl.col("__x_value").fill_null(0) >= 1)
        .select([pl.col(CASE_ID_COL), pl.col("x_bin"), pl.col(agg_col)])
    )

    y_melt = (
        df.select([CASE_ID_COL, *y_prefix_cols])
        .melt(
            id_vars=[CASE_ID_COL],
            value_vars=y_prefix_cols,
            variable_name="y_bin",
            value_name="__y_value",
        )
        .filter(pl.col("__y_value").fill_null(0) >= 1)
        .select([pl.col(CASE_ID_COL), pl.col("y_bin")])
    )

    temp_df = (
        x_melt.join(y_melt, on=CASE_ID_COL, how="inner")
        .select([pl.col(CASE_ID_COL), pl.col("x_bin"), pl.col("y_bin"), pl.col(agg_col)])
    )
    return temp_df


def apply(
    feature_table: pl.LazyFrame | pl.DataFrame,
    x_col: str,
    y_col: str,
    agg_col: str,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Tuple[pl.DataFrame, Dict[Any, Any]]:
    """Construct a process cube using Polars data structures."""

    if parameters is None:
        parameters = {}

    df = _ensure_polars_df(feature_table)
    if df.is_empty():
        return pl.DataFrame(), {}

    agg_fn_param = exec_utils.get_param_value(
        Parameters.AGGREGATION_FUNCTION, parameters, "mean"
    )
    agg_fn = _normalize_agg_fn(agg_fn_param)

    max_divisions_x = exec_utils.get_param_value(Parameters.MAX_DIVISIONS_X, parameters, 4)
    max_divisions_y = exec_utils.get_param_value(Parameters.MAX_DIVISIONS_Y, parameters, 4)
    x_bins_param = exec_utils.get_param_value(Parameters.X_BINS, parameters, None)
    y_bins_param = exec_utils.get_param_value(Parameters.Y_BINS, parameters, None)

    numeric_x = x_col in df.columns and _is_numeric_dtype(df.schema[x_col])
    numeric_y = y_col in df.columns and _is_numeric_dtype(df.schema[y_col])

    x_prefix_cols: List[str] = []
    y_prefix_cols: List[str] = []
    all_x_bins: List[str] = []
    all_y_bins: List[str] = []

    if not numeric_x:
        x_prefix_cols = _prefix_columns(df, x_col)
    if not numeric_y:
        y_prefix_cols = _prefix_columns(df, y_col)

    temp_df: pl.DataFrame

    if numeric_x and numeric_y:
        x_bins = _prepare_bins(df[x_col], x_bins_param, max_divisions_x)
        y_bins = _prepare_bins(df[y_col], y_bins_param, max_divisions_y)
        if len(x_bins) < 2 or len(y_bins) < 2:
            return pl.DataFrame(), {}
        all_x_bins = _bin_labels(x_bins)
        all_y_bins = _bin_labels(y_bins)
        temp_df = _numeric_numeric_case(df, x_col, y_col, agg_col, x_bins, y_bins)
    elif numeric_x and not numeric_y:
        x_bins = _prepare_bins(df[x_col], x_bins_param, max_divisions_x)
        if len(x_bins) < 2 or not y_prefix_cols:
            return pl.DataFrame(), {}
        all_x_bins = _bin_labels(x_bins)
        all_y_bins = y_prefix_cols
        temp_df = _numeric_prefix_case(df, x_col, agg_col, x_bins, y_prefix_cols, "y_bin")
    elif not numeric_x and numeric_y:
        y_bins = _prepare_bins(df[y_col], y_bins_param, max_divisions_y)
        if len(y_bins) < 2 or not x_prefix_cols:
            return pl.DataFrame(), {}
        all_x_bins = x_prefix_cols
        all_y_bins = _bin_labels(y_bins)
        temp_df = _prefix_numeric_case(df, y_col, agg_col, y_bins, x_prefix_cols, "x_bin")
    else:
        if not x_prefix_cols or not y_prefix_cols:
            return pl.DataFrame(), {}
        all_x_bins = x_prefix_cols
        all_y_bins = y_prefix_cols
        temp_df = _prefix_prefix_case(df, agg_col, x_prefix_cols, y_prefix_cols)

    if temp_df.is_empty():
        return pl.DataFrame(), {}

    temp_df = temp_df.filter(
        pl.col("x_bin").is_not_null() & pl.col("y_bin").is_not_null()
    )
    if temp_df.is_empty():
        return pl.DataFrame(), {}

    agg_expr = _aggregation_expression(agg_col, agg_fn).alias("value")
    grouped_values = (
        temp_df.group_by(["x_bin", "y_bin"], maintain_order=True)
        .agg(agg_expr)
    )

    pivot_df = grouped_values.pivot(
        values="value",
        index="x_bin",
        columns="y_bin",
        sort_columns=False,
    )

    if all_x_bins:
        base_rows = pl.DataFrame({"x_bin": all_x_bins})
        pivot_df = base_rows.join(pivot_df, on="x_bin", how="left")

    value_columns = [col for col in pivot_df.columns if col != "x_bin"]
    if all_y_bins:
        missing_cols = [col for col in all_y_bins if col not in value_columns]
        if missing_cols:
            additions = {col: [None] * pivot_df.height for col in missing_cols}
            pivot_df = pl.concat([pivot_df, pl.DataFrame(additions)], how="horizontal")
        ordered_cols = ["x_bin", *all_y_bins]
        remaining_cols = [col for col in pivot_df.columns if col not in ordered_cols]
        pivot_df = pivot_df.select([*ordered_cols, *remaining_cols])

    case_groups = (
        temp_df.group_by(["x_bin", "y_bin"], maintain_order=True)
        .agg(pl.col(CASE_ID_COL).unique().alias("case_ids"))
    )
    cell_case_dict = {
        (row[0], row[1]): set(row[2])
        for row in case_groups.iter_rows()
    }

    return pivot_df, cell_case_dict


__all__ = ["Parameters", "apply"]



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter
from math import log2
from typing import Dict, Union, Any, Optional, List
from pm4py.objects.log.obj import EventLog
import pandas as pd
from enum import Enum
from pm4py.util import constants, exec_utils, xes_constants


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    ALPHA = "alpha"


def apply(log: Union[pd.DataFrame, EventLog], parameters: Optional[Dict[Any, Any]] = None) -> List[Dict[str, Any]]:
    """
    Compute information–theoretic metrics used to detect *chaotic activities* in an event log, as defined in

    Tax, Niek, Natalia Sidorova, and Wil MP van der Aalst. "Discovering more precise process models from event logs by filtering out chaotic activities." Journal of Intelligent Information Systems 52.1 (2019): 107-139.

    The result maps each activity to:
    * freq                – absolute frequency #(a,L)
    * entropy             – H(a,L)  (direct entropy)
    * entropy_smooth      – Hₛ(a,L) (Laplace‑smoothed entropy)
    * entropy_gain        – ΔH      (drop in total log‑entropy if *a* is removed)
    * chaotic_score       – simple aggregate = (entropy_smooth+entropy_gain)/2

    Parameters
    -----------------
    log
        Event log or Pandas dataframe
    parameters
        Variant-specific parameters, including:
        - Parameters.ALPHA => Laplace/Lidstone smoothing parameter α. *None* reproduces the raw entropy H(a,L); a typical choice following the paper is α = 1/|A|.
        - Parameters.ACTIVITY_KEY => the attribute to be used as activity. Default: "concept:name"

    Returns
    -----------------
    chaotic_activities
        List of dictionaries, each representing an activity, sorted decreasingly based on the chaotic score (less is better).
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    alpha = exec_utils.get_param_value(Parameters.ALPHA, parameters, None)

    import pm4py

    traces = pm4py.project_on_event_attribute(log, activity_key)

    return chaotic_metrics(traces, alpha=alpha)


def chaotic_metrics(traces, alpha=None):
    """
    Parameters
    ----------
    traces   : list[list[str]]
        The event log where each inner list is a trace (ordered events).
    alpha : float | None
        Laplace/Lidstone smoothing parameter α.
        *None* reproduces the raw entropy H(a,L);
        a typical choice following the paper is α = 1/|A|.

    Returns
    -------
    dict[str, dict]   (activity  →  metrics)
    """
    # ------------------------------------------------------------------ pre‑compute counts
    act_count, pair_count = _pair_counts(traces)
    activities = set(act_count)
    if alpha is None:
        # default α = 1/|A|    (Sec. 3.2, last paragraph)
        alpha = 1.0 / len(activities)

    # ------------------------------------------------------------------ baseline totals
    H_total = total_entropy(traces, alpha=None)  # raw total entropy

    results = {}
    for a in sorted(activities):
        # ------------ H(a,L) and Hₛ(a,L)
        dfr_raw, dpr_raw = _entropy_vectors(a, activities,
                                            act_count, pair_count,
                                            alpha=None)
        dfr_sm, dpr_sm = _entropy_vectors(a, activities,
                                          act_count, pair_count,
                                          alpha=alpha)

        H_a = _entropy(dfr_raw) + _entropy(dpr_raw)
        Hs_a = _entropy(dfr_sm) + _entropy(dpr_sm)

        # ------------ indirect metric: ΔH = H(L) − H(L\{a})
        # Build log without *a*
        log_wo_a = [[x for x in t if x != a] for t in traces]
        # remove empty traces to avoid zero‑division
        log_wo_a = [t for t in log_wo_a if t]
        H_total_wo = total_entropy(log_wo_a, alpha=None)
        entropy_gain = H_total - H_total_wo  # larger ⇒ bigger reduction

        chaotic_score = (Hs_a + entropy_gain) / 2.0

        results[a] = {
            "freq": act_count[a],
            "entropy": H_a,
            "entropy_smooth": Hs_a,
            "entropy_gain": entropy_gain,
            "chaotic_score": chaotic_score
        }

    results = [{'activity': a, **data} for a, data in results.items()]
    results.sort(key=lambda x: (-x["chaotic_score"], -x["freq"], x["activity"]))

    return results


def _entropy(probabilities):
    """Shannon entropy of a discrete distribution (list/iterable of pᵢ)."""
    return -sum(p * log2(p) for p in probabilities if p > 0.0)


def _pair_counts(traces):
    """
    Return:
        act_count : Counter of single‑activity frequencies  #(a,L)
        pair_count: Counter of directly‑follows pairs       #( (a,b), L )
    """
    act_count = Counter()
    pair_count = Counter()

    for trace in traces:
        prev = constants.DEFAULT_ARTIFICIAL_START_ACTIVITY
        for act in trace:
            act_count[act] += 1
            pair_count[(prev, act)] += 1
            prev = act
        pair_count[(prev, constants.DEFAULT_ARTIFICIAL_END_ACTIVITY)] += 1
    return act_count, pair_count


def _entropy_vectors(act, activities, act_count, pair_count, alpha=None):
    """
    Build the directly‑precedes (dpr) and directly‑follows (dfr) probability
    vectors for *act*.  If alpha is given (≥0) Laplace/Lidstone smoothing
    is applied as in Eq. (1) of the paper.
    """
    k = len(activities) + 1  # +1 for END / START sentinel
    denom = act_count[act]
    if alpha is not None:  # Laplace denominator
        denom += alpha * k

    # --- directly‑follows -------------------------------------------------
    dfr = []
    for b in activities | {constants.DEFAULT_ARTIFICIAL_END_ACTIVITY}:
        num = pair_count.get((act, b), 0)
        p = (alpha + num) / denom if alpha is not None else num / denom
        dfr.append(p)

    # --- directly‑precedes ------------------------------------------------
    dpr = []
    for b in activities | {constants.DEFAULT_ARTIFICIAL_START_ACTIVITY}:
        num = pair_count.get((b, act), 0)
        # note: for precedes we smooth with same alpha and k
        p = (alpha + num) / denom if alpha is not None else num / denom
        dpr.append(p)

    return dfr, dpr


def total_entropy(traces, alpha=None):
    """Return Σₐ H(a,L) or Σₐ Hₛ(a,L)."""
    act_count, pair_count = _pair_counts(traces)
    activities = set(act_count)
    total = 0.0
    for a in activities:
        dfr, dpr = _entropy_vectors(a, activities, act_count, pair_count, alpha)
        total += _entropy(dfr) + _entropy(dpr)
    return total
